# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations


class Member:
    """
    Represents a Member of a Dimension. Members are immutable.
    Useful for building business logic and navigation through dimensions and data space.
    """
    _LEVEL = 6
    _NAME = 1

    def __init__(self, dimension, member_name, cube=None,
                 idx_dim: int = -1, idx_member: int = -1, member_level: int = -1):
        self._idx_dim = idx_dim
        self._idx_member = idx_member
        self._ordinal = dimension._member_idx_list.index(idx_member)
        self._member_level = member_level
        self._dimension = dimension
        self._name = member_name
        self._cube = cube

    def __repr__(self) -> str:
        return self._name

    def __str__(self) -> str:
        return self._name

    # region Properties
    @property
    def name(self) -> str:
        """Returns the name of the member."""
        return self._name

    @property
    def ordinal(self) -> int:
        """Returns the ordinal position of the member within the overall list of member."""
        return self._ordinal

    @property
    def full_name(self) -> str:
        """Returns the full qualified name of the member, incl. dimension"""
        return self._dimension.name + ":" + self._name

    @property
    def dimension(self):  # -> Dimension:
        """Returns the Dimension object the Member is associated to."""
        return self._dimension

    @property
    def cube(self):  # -> Cube:
        """Returns the Cube object the Member is associated to.
        If the Member object is not derived from a Cube or Cell, ``None``will be returned."""
        return self._cube

    @property
    def has_cube(self) -> bool:
        """Returns ``True`` if the Member object has been derived from a Cube or Cell context, and
        the Cube property will return an existing Cube instance. If ``False`` is returned, then the
        Cube property will return ``None```"""
        if self._cube:
            return True
        return False

    # endregion

    # region Navigation functions
    def __get_member(self, idx_member):
        member_level = self._dimension.members[idx_member][self._LEVEL]
        member_name = self._dimension.members[idx_member][self._NAME]
        return Member(self._dimension, member_name, self._cube, self._idx_dim, idx_member, member_level)

    def __update_member(self, idx_member):
        self._idx_member = idx_member
        self.member_level = self._dimension.members[idx_member][self._LEVEL]
        self.member_name = self._dimension.members[idx_member][self._NAME]

    def first(self) -> Member:
        """
        Returns the first Member of the dimension by its ordinal position.

        .. code:: python

            member = cube.dimensions("months").member("Jul")
            jan = member.first()  # 'Jan' is the first month defined in the months dimension.

        :return: A new Member object.
        """
        idx_member = list(self._dimension.members)[0]
        return self.__get_member(idx_member)

    def has_next(self) -> bool:
        """
        Check if a Member has a next Member, meaning it is not already the last member of the dimension.
        """
        return self._ordinal < self._dimension.member_counter - 1

    def next(self) -> Member:
        """
        Returns the next Member of the dimension by its ordinal position.

        .. code:: python

            member = cube.dimensions("months").member("Jul")
            aug = member.next()  # 'Aug' is the next month defined after 'Jul' in the months dimension.

        :return: A new Member object.
        """
        if self._ordinal < self._dimension.member_counter - 1:
            idx_member = self._dimension._member_idx_list[self._ordinal + 1]
            return self.__get_member(idx_member)
        raise IndexError(f"Member '{self.member_name}' is already the last member.")

    def has_previous(self) -> bool:
        """
        Check if a Member has a previous Member, meaning it is not already the first member of the dimension.
        """
        return self._ordinal > 0

    def previous(self) -> Member:
        """
        Returns the previous Member of the dimension by its ordinal position.

        .. code:: python

            member = cube.dimensions("months").member("Jul")
            jun = member.next()  # 'Jun' is the next month defined before 'Jul' in the months dimension.

        :return: A new Member object.
        """
        if self._ordinal > 0:
            idx_member = self._dimension._member_idx_list[self._ordinal - 1]
            return self.__get_member(idx_member)
        raise IndexError(f"Member '{self.member_name}' is already the first member.")

    def last(self) -> Member:
        """
        Returns the last Member of the dimension by its ordinal position.

        .. code:: python

            member = cube.dimensions("months").member("Jul")
            year = member.last()  # 'Year Total' is the last member defined in the months dimension.

        :return: A new Member object.
        """
        idx_member = list(self._dimension.members)[-1]
        return self.__get_member(idx_member)

    def has_parent(self) -> bool:
        """
        Check if a Member has at least one parent Member, meaning it is not already a top level member without parents.
        """
        return len(self._dimension.members[self._idx_member][self._dimension.PARENTS]) > 0

    def parent(self, index: int = 0) -> Member:
        """
        Returns the 1st or subsequent parent of a member. Equal to method ``up`` .

        .. code:: python

            member = cube.dimensions("months").member("Jul")
            q3 = member.up()  # 'Q3' is the first parent of 'Jul' in the months dimension.

        :type index: Index of the parent to return. 0 returns the first parent, 1 the second ...
        :raises KeyError: Raised, if a parent with the given index is not defined for the Member or if no parent exists.
        :return: A new Member object.
        """
        return self.up(index)

    def up(self, index: int = 0) -> Member:
        """
        Returns the 1st or subsequent parent of a member. Equal to method ``parent`` .

        .. code:: python

            member = cube.dimensions("months").member("Jul")
            q3 = member.up()  # 'Q3' is the first parent of 'Jul' in the months dimension.

        :type index: Index of the parent to return. 0 returns the first parent, 1 the second ...
        :raises KeyError: Raised, if a parent with the given index is not defined for the Member or if no parent exists.
        :return: A new Member object.
        """
        parents = self._dimension.members[self._idx_member][self._dimension.PARENTS]
        if 0 <= index < len(parents):
            return self.__get_member(parents[index])
        raise IndexError(f"Parent index {index} is out of range of available parents [0:{len(parents) - 1}] "
                         f"of member {str(self)}.")

    def child(self, index: int = 0) -> Member:
        """
        Returns the 1st or subsequent child of a member. Equal to method ``down`` .

        .. code:: python

            member = cube.dimensions("months").member("Q3")
            jul = member.down()  # 'Jul' is the first child of 'Q3' in the months dimension.
            aug = member.down(1)  # 'Aug' is the second child of 'Q3' in the months dimension.

        :type index: Index of the child to return. 0 returns the first child, 1 the second ...
        :raises KeyError: Raised, if a parent with the given index is not defined for the Member.
        :return: A new Member object.
        """
        return self.down(index)

    def down(self, index: int = 0) -> Member:
        """
        Returns the 1st or subsequent child of a member. Equal to method ``child`` .

        .. code:: python

            member = cube.dimensions("months").member("Q3")
            jul = member.down()  # 'Jul' is the first child of 'Q3' in the months dimension.
            aug = member.down(1)  # 'Aug' is the second child of 'Q3' in the months dimension.

        :type index: Index of the child to return. 0 returns the first child, 1 the second ...
        :raises KeyError: Raised, if a parent with the given index is not defined for the Member.
        :return: A new Member object.
        """
        children = self._dimension.members[self._idx_member][self._dimension.CHILDREN]
        if 0 <= index < len(children):
            return self.__get_member(children[index])
        raise IndexError(f"Child index {index} is out of range of available children [0:{len(children) - 1}] "
                         f"of member {str(self)}.")

    def is_root(self) -> bool:
        """
        Checks if a Member is a root member, meaning the member has no further parents.

        :return: ``True`` is the member is a root member, ``False`` otherwise.
        """
        return len(self._dimension.members[self._idx_member][self._dimension.PARENTS]) == 0

    def root(self, index: int = 0) -> Member:
        """
        Returns the 1st or subsequent root member of a member.

        .. code:: python

            member = cube.dimensions("months").member("Jul")
            year = member.root()  # 'Year Total' is the first root level parent of 'Jul' in the months dimension.

        :type index: Index of the root to return. 0 returns the first root, 1 the second ...
        :raises KeyError: Raised, if a root with the given index is not defined for the Member.
        :return: A new Member object.
        """
        roots = self._dimension.get_root_members()
        if 0 <= index < len(roots):
            return self.__get_member(roots[index])
        raise IndexError(f"Root index {index} is out of range [0:{len(roots) - 1}] of available root member "
                         f"for dimension {str(self._dimension.name)}.")

    def is_parent_of(self, other_member) -> bool:
        """
        Checks if a Member is a direct parent of another member.

        :return: ``True`` is the other member is a direct parent, ``False`` otherwise.
        """
        return NotImplemented

    def is_parent(self) -> bool:
        """
        Checks if a Member is a parent of some other member.

        :return: ``True`` is the member is a parent, ``False`` otherwise.
        """
        return self._dimension.members[self._idx_member][self._dimension.LEVEL] > 0

    def parents_count(self) -> int:
        """
        Returns the number of parents the Member has.

        :return: The number of parents the Member has.
        """
        return len(self._dimension.members[self._idx_member][self._dimension.PARENTS])

    def children(self) -> list[Member]:
        """
        Returns a list of all direct children of the member. If the member does not
        have children, then an empty array will be returned.

        :return: List of children.
        """
        return NotImplemented

    def base_members(self) -> list[Member]:
        """
        Returns a list of all base level members of the member. If the member is a base level member itself,
        meaning it does not have children, then an empty array will be returned.

        :return: List of base level members.
        """
        return NotImplemented

    def is_child_of(self, other_member) -> bool:
        """
        Checks if a Member is a direct child of another member.

        :return: ``True`` is the other member is a direct child, ``False`` otherwise.
        """
        return NotImplemented

    def is_child(self) -> bool:
        """
        Checks if a Member is a child of some other member.

        :return: ``True`` is the member is a child, ``False`` otherwise.
        """
        return len(self._dimension.members[self._idx_member][self._dimension.PARENTS]) > 0

    def children_count(self) -> int:
        """
        Returns the number of children the Member has.

        :return: The number of children the Member has.
        """
        return len(self._dimension.members[self._idx_member][self._dimension.CHILDREN])

    def has_children(self) -> bool:
        """
        Checks if a Member has at least one child.

        :return: ``True`` is the member has at least one child, ``False`` otherwise.
        """
        return self._member_level > 0

    def parents(self) -> list[Member]:
        """
        Returns a list of all direct parents of the member. If the member does not
        have parents, then an empty array will be returned.

        :return: List of parents.
        """
        return NotImplemented

    def level(self) -> int:
        """
        Returns the level of member. 0 indicates base level members, higher values aggregated members.

        :return: Level of the member.
        """
        return self._member_level

    def is_base_member(self) -> bool:
        """
        Checks if a Member is a base level member.

        :return: ``True`` is the member is a base level member, ``False`` otherwise.
        """
        return self._member_level == 0

    def is_aggregated_member(self) -> bool:
        """
        Checks if a Member is an aggregated member.

        :return: ``True`` is the member is an aggregated member, ``False`` otherwise.
        """
        return self._member_level > 0

    # endregion



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import os.path
from collections.abc import Iterable
from copy import deepcopy
from typing import Tuple

import tinyolap.utils
from tinyolap.storage.sqlite import SqliteStorage
from tinyolap.storage.storageprovider import StorageProvider
from tinyolap.encryption import EncryptionMethodEnum, Encryptor, NotAnEncryptor, ObfuscationEncryptor, FernetEncryptor
from tinyolap.case_insensitive_dict import CaseInsensitiveDict
from tinyolap.codemanager import CodeManager
from tinyolap.cube import Cube
from tinyolap.dimension import Dimension
from tinyolap.exceptions import *
from tinyolap.history import History


class Database:
    """
    Databases are the root objects that should be use to create **TinyOlap** data model or database.
    The Database object provides methods to create :ref:`dimensions <dimensions>` and :ref:`cubes <cubes>`.
    """
    MIN_PASSWORD_LENGTH = 4
    MIN_DIMS_PER_CUBE = 1
    MAX_DIMS_PER_CUBE = 32  # This value can be changed.
    # Note: 32 dimensions is already huge for a model-driven OLAP database.
    MAX_MEASURES_PER_CUBE = 1024  # Value can be changed: max. measures = (2000 - MAX_DIMS_PER_CUBE)

    def __init__(self, name: str = None, in_memory: bool = True,
                 encryption: EncryptionMethodEnum = EncryptionMethodEnum.NoEnryption,
                 password: str = None):
        """
        Creates a new database or, if a database file with the same name already exists, opens an existing database.

        The opening of an existing database will restore the state of the database either before the
        last call of the ``.close()`` method or before the Database object was released the last time.

        If a database file exists for the given database name and parameter ``ìn-memory`` will be set to ``True`,
        then the existing database file will not be opened, changed or overwritten.

        :param name: Name of the database. Only alphanumeric characters and underscore are supported for database names
        (no whitespaces or special characters).

        :param in_memory: Identifies if the database should run in memory only (no persistence) or should persist
        all changes to disk. If `ìn-memory``wil be set to ``True`, then a potentially existing database file for the
        given database name will not be opened, changed or overwritten. To save a database running in in memory mode,
        use the ``save()``method of the database object.

        :param encryption: Encryption method for database contents when ``ìn-memory = True``.
        By default no encryption is used, ``encryption = EncryptionMethodEnum.NoEnryption``.

        .. note::
            If encryption is required, using ``encryption = EncryptionMethodEnum.Obfuscation`` should
            be sufficient for most cases and is recommended for performance reasons.
            Use ``encryption = EncryptionMethodEnum.Encryption`` only if the highest possible security
            level is required. For such cases also ensure to choose a long (>= 32 charactes) and cryptic
            password consisting of upper- and lower-case chatacter, number and special characters.

        :param password: Password for encryption. Require if any other encryption method is required than
        ``encryption = EncryptionMethodEnum.NoEnryption``. Please ensure to remember the password.
        """
        self._file_name = None
        if name != tinyolap.utils.to_valid_key(name):
            if os.path.exists(name):
                self._file_name = name
            else:
                raise InvalidKeyException(f"'{name}' is not a valid database name or path. "
                                          f"For database names alphanumeric characters and underscore supported only, "
                                          f"no whitespaces, no special characters.")
        self.dimensions: CaseInsensitiveDict[str, Dimension] = CaseInsensitiveDict()
        self.cubes: CaseInsensitiveDict[str, Cube] = CaseInsensitiveDict()
        self._code_manager: CodeManager = CodeManager()
        self._history: History = History(self)
        self._name: str = name
        self._file_name: str = name

        self._in_memory = in_memory
        if in_memory:
            self._encryptor: Encryptor = NotAnEncryptor()
            self._storage_provider: StorageProvider = None
        else:
            # setup encryption
            if encryption == EncryptionMethodEnum.NoEnryption:
                self._encryptor: Encryptor = NotAnEncryptor()
            else:
                if not password or len(password.strip()) < self.MIN_PASSWORD_LENGTH:
                    raise ValueError(f"Failed to activate encryption. The defined password does not meet "
                                     f"the minimum requirements of: length >= {self.MIN_PASSWORD_LENGTH}.")
                if encryption == EncryptionMethodEnum.Obfuscation:
                    self._encryptor: Encryptor = ObfuscationEncryptor(password)
                elif encryption == EncryptionMethodEnum.Encryption:
                    self._encryptor: Encryptor = FernetEncryptor(password)
                else:
                    raise ValueError(f"Failed to activate encryption. '{encryption}' is not a "
                                     f"supported encryption method.")

            # setup the storage provider
            self._storage_provider: StorageProvider = SqliteStorage(name=self._name, encryptor=self._encryptor)
            self._storage_provider.open(file_name=self._file_name)
        self._load()
        self._caching = True

    def __del__(self):
        self.close()

    # region Properties
    @property
    def in_memory(self) -> bool:
        """
        Identifies if the database is in-memory mode. Please note that the in-memory
        property can not be changed after the initialization of a database object.

        :return: Returns ``True`` if the database is in-memory mode, ``False`` otherwise.
        """
        return self._in_memory

    @property
    def history(self) -> History:
        """Returns the history of the database."""
        return self._history

    @property
    def code_manager(self) -> CodeManager:
        """Returns the code manager of the database."""
        return self._code_manager

    @property
    def name(self) -> str:
        """Returns the name of the database."""
        return self._name

    @property
    def uri(self) -> str:
        """Returns the uri of the database."""
        if self._storage_provider:
            return self._storage_provider.uri
        return None

    @property
    def file_path(self) -> str:
        """Returns the file path of the database."""
        if self._storage_provider:
            file = self._storage_provider.uri
            if file.startswith("file://"):
                file = file[7:]
            return file
        return None

    @property
    def caching(self) -> bool:
        """
        Identifies if caching is activated for the database.
        By default, caching is activated for all :ref:`cubes <cubes>`.
        The caching setting for individual cubes can be overwritten.

        .. note::
            It is highly recommended, especially for larger data models containing
            a lot of data maybe even essential, to always set caching to ``True```
            what anyhow is the default value. Caching greatly improves the user
            experience by caching values for **aggregated cells** from cubes.

            Please be aware that the cache needs to be warmed up on first use.
            For future versions of **TinyOlap** it is planned to also persist the
            cube cache to enable faster initial cell access.

        :return: Returns ``True`` if caching is activated for the database, ``False`` otherwise.
        """
        return self._caching

    @caching.setter
    def caching(self, value: bool):
        """
        Identifies if caching is activated for the database.
        By default, caching is activated for all :ref:`cubes <cubes>`.
        The caching setting for individual cubes can be overwritten.

        .. note::
            It is highly recommended, especially for larger data models containing
            a lot of data maybe even essential, to always set caching to ``True```
            what anyhow is the default value. Caching greatly improves the user
            experience by caching values for **aggregated cells** from cubes.

            Please be aware that the cache needs to be warmed up on first use.
            For future versions of **TinyOlap** it is planned to also persist the
            cube cache to enable faster initial cell access.

        :param value: Set value to ``True`` to activate caching, ``False`` to deactivate caching.
        """
        self._caching = value
        for cube in self.cubes.values():
            cube.caching = value

    # endregion

    # region Database related methods
    def rename(self, new_name: str):
        """Renames the database.
        :param new_name: New name for the database.
        :raise KeyError: Raised if the key is invalid.
        """
        if new_name != tinyolap.utils.to_valid_key(new_name):
            raise KeyError(f"'{new_name}' is not a valid database name. "
                           f"alphanumeric characters and underscore supported only, "
                           f"no whitespaces, no special characters.")
        self._name: str = new_name

        if self._storage_provider:
            self._storage_provider.add_meta("db", self._to_json())

    def clone(self):
        """Creates a clone of the database."""
        return deepcopy(self)

    def open(self, file_name: str):
        """
        Opens a database from a file.

        .. Note::
            Please note that an already defined database name will be overwritten by
            the name defined in the dabase file.

        .. caution::
            If the database is set to ``in_memory == True``, then the requested database file
            will be opened and used only to initialize the database (like a template).
            But it will NOT be connected and write back change to that database file.
            So, any subsequent change, after opening a database file, will happen in-memory
            only and not be persisted. Please remember, that you can use the ``export()``
            method to save an in-memory databases at any time.

        :param file_name: The database file to open.
        """
        self._load(file_name)

    def save(self):
        """
        Saves pending changes to the database. The ``save()`` command has no effect
        if the database is in-memory mode, meaning ``in_memory == True``.

        :return:
        """
        if self._storage_provider:
            self._storage_provider.add_meta("db", self._to_json())
            self._storage_provider.add_meta("code", self._code_manager.to_json())
            for cube in self.cubes.values():
                self._storage_provider.add_cube(cube.name, cube.to_json())

    def close(self):
        """
        Closes the database. If in-memory mode is off ``in_memory == False`` ,
        then all pending changes to the database file will be committed and
        the connection to the underlying database will be be closed.

        The ``close()`` command has no effect if the database is in-memory mode,
        meaning ``in_memory == True``.
        """
        if self._storage_provider:
            if self._code_manager.pending_changes:
                self._storage_provider.add_meta("code", self._code_manager.to_json())
                self._code_manager.pending_changes = False
            self._storage_provider.close()

    def delete(self):
        """
        Deletes the database file and log file, if such exists. If the database is
        open, it will be closed beforehand. You should handle this method with care.
        The ``delete()`` command has no effect if the database is in in-memory mode
        ``in_memory == True``.
        """
        if self._storage_provider:
            self._storage_provider.close()
            self._storage_provider.delete()

    def export(self, name: str, overwrite_if_exists: bool = False,
               encryption: EncryptionMethodEnum = EncryptionMethodEnum.NoEnryption,
               password: str = None):
        """
        Exports the database to a new database file. This method is useful e.g.
        for creating backups and especially to persist databases that run in
        in-memory mode.

        .. note::
            When **TinyOlap** is used for data processing purposes, rather than
            as for planning or other data-entry focussed tasks, it is a clever
            idea to spin up the database first in in-memory mode, do all your
            processing and then persists the result using the ``export(...)``
            method. This approach is presumably much faster than constantly
            writing to the database file.

        :param name: Either a simple name or a fully qualified file path. If
           the name does not represent a path, then the database file will be
           created in the default location */db* or whatever is specified in
           the config file for ``database_folder``.

        :param overwrite_if_exists: Defines if an already existing file should
           be overwritten or not. If set to ``False`` and the file already exist,
           an FileExistsError will be raised.

        :param in_memory: Identifies if the database should run in memory only (no persistence) or should persist
           all changes to disk. If `ìn-memory``wil be set to ``True`, then a potentially existing database file for the
           given database name will not be opened, changed or overwritten. To save a database running in in memory mode,
           use the ``save()``method of the database object.

        :param encryption: Encryption method for database contents when ``ìn-memory = True``.
          By default no encryption is used, ``encryption = EncryptionMethodEnum.NoEnryption``.

        .. note::
            If encryption is required, using ``encryption = EncryptionMethodEnum.Obfuscation`` should
            be sufficient for most cases and is recommended for performance reasons.
            Use ``encryption = EncryptionMethodEnum.Encryption`` only if the highest possible security
            level is required. For such cases also ensure to choose a long (>= 32 charactes) and cryptic
            password consisting of upper- and lower-case chatacter, number and special characters.

        :param password: Password for encryption. Require if any other encryption method is required than
           ``encryption = EncryptionMethodEnum.NoEnryption``. Please ensure to remember the password.

        :return:
        """
        if name.lower() == self.name.lower():
            raise DatabaseBackendException(f"Failed to export database '{self.name}'. "
                                           f"You cannot export a database under it's current name.")

        exporter: StorageProvider = SqliteStorage(name)
        if exporter.exists():
            if not overwrite_if_exists:
                raise FileExistsError(f"Failed to export database '{self.name}'. "
                                      f"The database file '{exporter.uri}' already exists.")
            else:
                exporter.delete()

        # Export the database
        exporter.open()
        exporter.add_meta("db", self._to_json())
        exporter.add_meta("code", self._code_manager.to_json())
        for dimension in self.dimensions.values():
            exporter.add_dimension(dimension.name, dimension.to_json())
        for cube in self.cubes.values():
            exporter.add_cube(cube.name, cube.to_json())
            exporter.set_records(cube.name, cube._get_records())
        exporter.close()

    def _to_json(self) -> str:
        """
        Returns the database configuration as a json string.
        :return: A json string.
        """
        config = {"name": self._name,
                  "caching": self._caching,
                  "dimensions": len(self.dimensions),
                  "cubes": len(self.cubes),
                  "contains_code": len(self._code_manager.functions) > 0,
                  "max_dims_per_cube": self.MAX_DIMS_PER_CUBE,
                  "max_measures_per_cube": self.MAX_MEASURES_PER_CUBE}
        return json.dumps(config)

    def _from_json(self, data: str):
        """
        Initializes the database meta data from a configuration in json format.
        :return: A json string.
        """
        if data:
            config = json.loads(data)
            self._name = config["name"]
            self._caching = config["caching"]

    # endregion

    # region Cell access via indexing
    def __getitem__(self, item):
        cube = self.cubes[item[0]]
        return cube.get(item[1:])

    def __setitem__(self, item, value):
        cube = self.cubes[item[0]]
        cube.set(item[1:], value)

    def __delitem__(self, item):
        cube = self.cubes[item[0]]
        cube.set(item[1:], None)

    # endregion

    # region Dimension related methods
    def add_dimension(self, name: str, description: str = None) -> Dimension:
        """Adds a new :ref:`dimension <dimensions>` to the database.

        :param name: Name of the dimension to be added.
        :param description: Description for the dimension.
        :return Dimension: The newly added dimension.
        :raises InvalidDimensionNameException: If the dimension name is invalid.
        :raises DuplicateDimensionException: If a dimension with the same name already exists.
        """
        if not tinyolap.utils.is_valid_db_object_name(name):
            raise InvalidKeyException(f"'{name}' is not a valid dimension name. "
                                      f"Lower case alphanumeric characters and underscore supported only, "
                                      f"no whitespaces, no special characters.")
        if name in self.dimensions:
            raise DuplicateKeyException(f"Failed to add dimension. A dimension named '{name}' already exists.")
        dimension = Dimension._create(self._storage_provider, name, description=description)
        dimension.database = self
        self.dimensions[name] = dimension
        return dimension

    def dimension_remove(self, dimension):
        """Removes a :ref:´dimension <dimensions>´ from the database.

        :param dimension: Name of the dimension, or the :ref:`dimension <dimensions>` object to be removed.
        :raises KeyNotFoundError: If the dimension not exists.
        """
        if type(dimension) is str:
            name = dimension
        else:
            name = dimension._name
        if name not in self.dimensions:
            raise KeyNotFoundError(f"A dimension named '{name}' does not exist.")

        uses = [cube.name for cube in self.cubes.values() if len([name in [dim.name for dim in cube._dimensions]])]
        if uses:
            raise DimensionInUseException(f"Dimension '{name}' is in use by cubes ({', '.join(uses)}) "
                                          f"and therefore can not be removed. Remove cubes first.")

        if self._storage_provider and self._storage_provider.connected:
            self._storage_provider.remove_dimension(name)

        del self.dimensions[name]

    def dimension_exists(self, name: str):
        """Checks if a dimension exists.

        :param name: Name of the dimension to be checked.
        :returns bool: Returns ``True`` if the dimension exists, ``False`` otherwise."""
        return name in self.dimensions

    # endregion

    # region Cube related methods
    def add_cube(self, name: str, dimensions: list, measures=None, description: str = None):
        """
        Creates a new :ref:`cube<cubes>` and adds it to the database.

        :param name: Name of the cube to be created.
        :param dimensions: A list of either names of existing dimensions of the database
           or :ref:`dimension <dimensions>` objects contained in the database.
        :param measures: (optional) a measure name or a list of measures names for the cube.
           If argument 'measures' is not defined, that a default measure named 'value' will be created.
        :param description: (optional) description for the cube.
        :return: The added cube object.
        :raises CubeCreationException: Raised if the creation of the cubed failed due to one
           of the following reasons:

        * The cube name is not invalid. Cube have to consist of lower case alphanumeric characters
            or underscore only, blanks or special characters are not allowed.

        * The list of dimensions is empty, contains dimension names or objects not associated with
          the cube, or the number of dimensions exceeds the upper limit of dimensions supported
          for cube creation.

          .. note::
             The default upper limit for cube creation is 32 dimensions, but this limit can be
             adjusted at any time by changing the value for ``MAX_DIMS_PER_CUBE`` in the source file
             'database.py'.

        :raises DuplicateKeyException: Raised if the cube already exists.")
        """

        # validate cube name
        if not tinyolap.utils.is_valid_db_object_name(name):
            raise CubeCreationException(f"Invalid cube name '{name}'. Cube names must contain "
                                        f"lower case alphanumeric characters only, no blanks or special characters.")
        if name in self.cubes:
            raise DuplicateKeyException(f"A cube named '{name}' already exists.")

        # validate dimensions
        if not dimensions:
            raise CubeCreationException("List of dimensions to create cube is empty or undefined.")
        if len(dimensions) > self.MAX_DIMS_PER_CUBE:
            raise CubeCreationException(f"Too many dimensions ({len(dimensions)}). "
                                        f"Maximum number dimensions per cube is {self.MAX_DIMS_PER_CUBE}.")
        dims = []
        for dimension in dimensions:
            if type(dimension) is str:
                if dimension not in self.dimensions:
                    raise CubeCreationException(f"A dimension named '{str(dimension)}' is not defined in "
                                                f"database '{self.name}'.")
                dims.append(self.dimensions[dimension])
            elif type(dimension) is Dimension:
                if dimension.name not in self.dimensions:
                    raise CubeCreationException(f"Dimension '{str(dimension.name)}' is not defined in "
                                                f"database '{self.name}'.")
                dim = self.dimensions[dimension.name]
                if dim is not dimension:
                    raise CubeCreationException(f"Dimension '{str(dimension.name)}' is not the same dimension "
                                                f"as the one defined in database '{self.name}'. You can only use "
                                                f"dimensions from within a database to add cubes.")

                dims.append(dimension)
            else:
                raise CubeCreationException(f"Unsupported dimension type '{str(dimension)}'.")
        # validate measures
        if measures:
            if type(measures) is str:
                if not tinyolap.utils.is_valid_member_name(measures):
                    raise CubeCreationException(f"Measure name '{str(measures)}' is not a valid measure name. "
                                                f"Please refer the documentation for further details.")
            elif isinstance(measures, Iterable):
                for m in measures:
                    if not tinyolap.utils.is_valid_member_name(m):
                        raise CubeCreationException(f"Measure name '{str(m)}' is not a valid measure name. "
                                                    f"Please refer the documentation for further details.")
        # create and return the cube
        cube = Cube.create(self._storage_provider, name, dims, measures, description)
        cube.caching = self.caching
        cube._database = self
        self.cubes[name] = cube
        return cube

    def cube_exists(self, name: str):
        """Checks if a cube exists.

        :param name: Name of the cube to be checked.
        :returns bool: Returns ``True`` if the cube exists, ``False`` otherwise."""
        return name in self.cubes

    def set(self, cube: str, address: Tuple[str], value: float):
        """
        Writes a value to the database for the given cube, idx_address and measure.
        Write back is supported for base level cells only.

        .. note:: Although TinyOlap is intended to be used for numerical data mainly,
            any Python data type or object reference can be written to a database.
            Persistence is only provided/guarantied for build-in base data types.
            If you want to serialized/deserialized custom objects through cube cells,
            then write/read json and do serialization and deserialization by yourself.

        :param cube: Name of the cube to write to.
        :param address: Address of the cube cell to write to.
        :param value: The value to be written to the database.
        """
        self.cubes[cube].set(address, value)

    def get(self, cube: str, address: Tuple[str], measure: str):
        """Returns a value from the database for a given cube, idx_address and measure.
                If no records exist for a given valid idx_address, then 0.0 will be returned."""
        return self.cubes[cube].get(address, measure)

    # endregion

    # region internal functions
    def _load(self, file_name: str = None):
        """Initialize database from storage storage_provider."""

        if self._storage_provider:
            provider = self._storage_provider
            if not provider.exists():
                return
            if not provider.connected:
                if file_name:
                    provider.open(file_name=file_name)
                else:
                    provider.open()

            # initialize meta information
            self._from_json(provider.get_meta("db"))
            self._code_manager.from_json(provider.get_meta("code"))

            # initialize dimensions
            data = provider.get_dimensions()
            for dim_tuple in data:
                dim_name, dim_json = dim_tuple
                dimension = self.add_dimension(dim_name)
                dimension.from_json(dim_json)

            # initialize cubes
            data = provider.get_cubes()
            for cube_tuple in data:
                cube_name, cube_json = cube_tuple
                cube = Cube.create(provider, cube_name, None, None, None)
                cube._database = self
                cube.from_json(cube_json)
                self.cubes[cube_name] = cube

            # import data
            # todo: implementation missing

    def _remove_members(self, dimension, members):
        """Remove data for obsolete (deleted) members over all cubes.
        Formulas containing that member will get invalidated."""

        # broadcast to all cubes...
        for cube in self.cubes.values():
            cube._remove_members(dimension, members)

    def _flush_cache(self):
        """Flushes all caches of all cubes"""
        for cube in self.cubes.values():
            cube._cache = {}

    # endregion



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from enum import Enum
from pathlib import Path
from timeit import default_timer as timer

import utils


class Logger:

    LOG_FILE_EXTENSION = ".log"

    def __init__(self, name: str = "tinyolap", level=logging.INFO):
        name = name.lower().strip()
        if not name.startswith("tinyolap."):
            name = "tinyolap." + name
        self.name = name
        self.root = name.split(".")[0]
        self.level = level
        self.file_path = utils.get_file_path(self.root + self.LOG_FILE_EXTENSION)

        self.logger = logging.getLogger(self.name)
        handler = logging.FileHandler(self.file_path, mode='w')
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(self.level)


    def __setup_logger(self):
        self.logger = logging.getLogger("storage_provider")
        handler = logging.FileHandler(self.file_path, mode='w')
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(self.LOG_LEVEL)

    def delete_log_file(self):
        """Deletes the database log file."""
        try:
            if os.path.exists(self.log_file):
                os.remove(self.log_file)
            return True
        except OSError:
            return False


from __future__ import annotations
import datetime
from abc import ABC, abstractmethod
import logging
import os
import string
import sqlite3
from collections.abc import Iterable
from os import path
from pathlib import Path
from timeit import default_timer as timer
import tinyolap.utils
from tinyolap.storage.storageprovider import StorageProvider
from tinyolap.exceptions import *
from tinyolap.encryption import Encryptor, NotAnEncryptor


class SqliteStorage(StorageProvider):
    """SQLite 3 storage provider. Default for persistence in TinyOlap."""

    LOG_PREFIX = " SqliteStorage - "

    # file related names
    DB_DEFAULT_FOLDER_NAME = "db"
    DB_EXTENSION = ".tinyolap.db"
    LOG_EXTENSION = ".tinyolap.log"
    LOG_LEVEL = logging.INFO

    # database object names
    META_TABLE_META = "tinyolap_meta"
    META_TABLE_CUBES = "tinyolap_cubes"
    META_TABLE_DIMENSIONS = "tinyolap_dimensions"
    META_TABLE_FIELDS = [("key", "TEXT PRIMARY KEY"), ("config", "TEXT")]

    HISTORY_TABLE = "tinyolap_history"
    HISTORY_TABLE_FIELDS = [("timestamp", "TEXT"), ("user", "TEXT"), ("action", "TEXT"), ("data", "TEXT")]

    DATA_TABLE_PREFIX = "tinyolap_data_"

    def __init__(self, name: str, database_folder: str = None, enable_logging: bool = True,
                 logger: logging.Logger = None, encryptor: Encryptor = None):
        self.name = name.strip()
        self.database_folder = database_folder
        self.logging = enable_logging
        self.logger = logger
        self.log_file = None
        if self.logging and not self.logger:
            self._initialize_logger()
        if encryptor is None:
            encryptor = NotAnEncryptor()
        self.encryptor = encryptor
        self.file_name = None
        self.folder = None
        self.file_path = None
        self.is_open = False
        self.conn: sqlite3.Connection = None
        self.cursor: sqlite3.Cursor = None

    # region basic database handling - open , close, delete, ...
    @property
    def connected(self) -> bool:
        """
        Identifies if the storage provider is connected or open.
        :return: ``True`` if the storage provider is connected, ``False`` otherwise.
        """
        return self.is_open

    @property
    def uri(self) -> str:
        """
        Returns or sets the uri (uniform resource identifier) of the current database.
        :return: The uri of the database.
        """
        if not self.file_path:
            file_exists, folder, self.file_path, self.file_name = self._evaluate_path(self.name)
            if file_exists:
                return self.file_path.as_uri()
        return self.file_path.as_uri()

    @uri.setter
    def uri(self, value: str):
        """
        Returns or sets the uri (uniform resource identifier) of the current database.
        :return: The uri of the database.
        """
        file_exists, self.folder, self.file_path, self.file_name = self._evaluate_path(value)

    def open(self, **kwargs) -> bool:
        """
        Opens the database. If the database does not exist, a new database file will be created.
        :param kwargs:
        :return: Returns ``True``if the database has been opened successfully.
        :raises DatabaseBackendException: Raised when the opening of / connecting to the database file failed.
        """
        if kwargs:
            file_name = kwargs["file_name"]
            if file_name:
                self.name = file_name

        file_exists, self.folder, self.file_path, self.file_name = self._evaluate_path(self.name)
        self.log_file = str(self.file_path) + self.LOG_EXTENSION
        self._initialize_logger()

        self.close()
        try:
            if self.logging:
                self.logger.info(f"{self.LOG_PREFIX}Opening database '{str(self.file_path)}'")
                # self.logger.handlers[0].flush()
            self.conn: sqlite3.Connection = sqlite3.connect(self.file_path)
            self.cursor: sqlite3.Cursor = self.conn.cursor()

            self._prepare_database()
            self.is_open = True
        except sqlite3.Error as err:
            msg = f"{self.LOG_PREFIX}Failed to open database '{self.file_path}'. SQLite exception: {str(err)}"
            if self.logging:
                self.logger.error(msg)
            raise DatabaseBackendException(msg)
        except Exception as err:
            msg = f"{self.LOG_PREFIX}Failed to open database '{self.file_path}'. {str(err)}"
            if self.logging:
                self.logger.error(msg)
            raise DatabaseBackendException(msg)
        return True

    def close(self) -> bool:
        """
        Closes the database.
        :return: ``True`` if the database was successfully closed
            or was already closed, ``False``otherwise.
        """
        if self.is_open:
            if self.logging and self.logger:
                self.logger.info(f"{self.LOG_PREFIX}Closing database '{self.file_path}'.")
                # self.logger.handlers[0].flush()
            if self.conn or self.is_open:
                try:
                    self._commit()
                    self.conn.close()
                except sqlite3.Error as err:
                    msg = f"{self.LOG_PREFIX}Failed to close database '{self.file_path}'. SQLite exception: {str(err)}"
                    if self.logging:
                        self.logger.error(msg)
                        # self.logger.handlers[0].flush()
                    raise DatabaseBackendException(msg)
                except Exception as err:
                    msg = f"{self.LOG_PREFIX}Failed to close database '{self.file_path}'. {str(err)}"
                    if self.logging:
                        self.logger.error(msg)
                        # self.logger.handlers[0].flush()
                    raise DatabaseBackendException(msg)

        self.is_open = False
        return True

    def exists(self) -> bool:
        """
        Checks if a database with the given name exists.
        :return: ``True`` if the database file exists, ``False``otherwise.
        """
        file_exists, folder, file_path, file_name = self._evaluate_path(self.name)
        return file_exists

    def delete(self) -> bool:
        """
        Deletes the database file and the database log file, if such exist.
        :return: True if the database file and the database log file was deleted successfully
        or if both do not exist.
        """
        file_exists, folder, file_path, file_name = self._evaluate_path(self.name)
        try:
            if self.is_open:
                self.close()
            if path.exists(file_path):
                if self.logging and self.logger:
                    self.logger.info(f"{self.LOG_PREFIX}Deleting database file '{self.file_path}'.")
                    # self.logger.handlers[0].flush()
                os.remove(file_path)
            self.delete_log()
            return True
        except OSError as err:
            if self.logging:
                self.logger.error(f"{self.LOG_PREFIX}Failed to delete database file '{self.file_path}'. {str(err)}")
                # self.logger.handlers[0].flush()
            return False

    def delete_log(self) -> bool:
        """
        Deletes the database log file, if such exists.
        :return: True if the log file was deleted successfully or if the log file not exist.
        """
        file_exists, folder, file_path, file_name = self._evaluate_path(self.name)
        self.log_file = str(file_path) + self.LOG_EXTENSION
        try:
            if path.exists(self.log_file):
                os.remove(self.log_file)
            return True
        except OSError:
            return False

    # endregion

    # region History related methods
    def add_to_history(self, time_stamp: datetime.datetime, user: str, action: str, data: str):
        """
        Add an event to the database history.
        :param time_stamp: Timestamp of the event
        :param user: The name of the user that initiated the event.
        :param action: The name of action executed through the event.
        :param data: The data related to the event. Defines what and how to undo or redo the event.
        """
        # todo: implementation missing
        raise NotImplementedError()

    def add_many_to_history(self, records: list[tuple[datetime.datetime, str, str, str]]):
        """
        Add multiple events to the database history.
        :param records: A list of tuples of format (time_stamp:datetime, user:str, action:str, data:str)
        """
        # todo: implementation missing
        raise NotImplementedError()

    def clear_history(self):
        """
        Clears the entire history.
        """
        # todo: implementation missing
        raise NotImplementedError()

    def delete_history_earlier_than(self, time_stamp: datetime.datetime):
        """
        Deletes all events from the history earlier than a specific timestamp.
        :param time_stamp: The timestamp to apply.
        """
        # todo: implementation missing
        raise NotImplementedError()

    def delete_history_later_than(self, time_stamp: datetime.datetime):
        """
        Deletes all events from the history later than a specific timestamp.
        :param time_stamp: The timestamp to apply.
        """
        # todo: implementation missing
        raise NotImplementedError()

    def delete_history_by_rowid(self, rowid):
        """
        Deletes events from the history by their rowid. Argument ``rowids`` can either be a single
        int value or an iterable of int.
        :param rowid: Either a single int value, or an iterable of int.
        """
        # todo: implementation missing
        raise NotImplementedError()

    def get_history_from_time_window(self, from_time_stamp: datetime.datetime = datetime.datetime.min,
                                     to_time_stamp: datetime.datetime = datetime.datetime.max) \
            -> list[tuple[datetime.datetime, str, str, str, int]]:
        """
        Returns the history records for a given time window.
        :param from_time_stamp: Begin of the time window.
        :param to_time_stamp: End of the time window.
        :return: A list of tuples of format (time_stamp:datetime, user:str, action:str, data:str, rowid:int)
        sorted in ascending order by the timestamp.
        """
        # todo: implementation missing
        raise NotImplementedError()

    def get_history(self, count: int = 1) -> list[tuple[datetime.datetime, str, str, str, int]]:
        """
        Returns the a certain number of records from the history.
        :param count: Number of records to be returned.
        :return: A list of tuples of format (time_stamp:datetime, user:str, action:str, data:str, rowid:int)
        sorted in ascending order by the timestamp.
        """
        a = f"SELECT * FROM ContentMaster WHERE ContentAddedByUserID = '%@' ORDER BY rowid DESC LIMIT {count}"
        # todo: implementation missing
        raise NotImplementedError()

    def count_history(self):
        """
        Returns the number of records in the history.
        """
        # todo: implementation missing
        raise NotImplementedError()

    # endregion

    # region Cell (data record) related methods
    def set_record(self, cube_name, address: str, data: str = None, instant_commit: bool = True) -> bool:
        """
        Add, updates or deletes a record related to a specific address and cube.
        To delete data, set argument ``data`` to ``None``.
        :param cube_name: Name of the targeted cube.
        :param address: The address of the cube cell as a string.
          Recommendation: use *str(idx_address:tuple[int])* to generate address from int cell address tuples.
        :param data: The data to be set as a string. The json format should be applied, or any string.
        :param instant_commit: If set to ``True``, then the change is instantly committed to the underlying database.
        :return: ``True`` if successful.
        """
        # todo: add encryption

        table = self.DATA_TABLE_PREFIX + cube_name
        if not data:
            sql = f"DELETE FROM {table} WHERE address = '{address}';"
            self._execute(sql)
        else:
            sql = f"INSERT INTO {table}(address, data) VALUES(?,?) " \
                  f"ON CONFLICT(address) DO UPDATE SET data=EXCLUDED.data;"
            if self.encryptor:
                data = self.encryptor.encrypt(data)
            self._execute(sql, (address, data))
        if instant_commit:
            self._commit()
        return True

    def set_records(self, cube_name, records: list[tuple[str, str]], instant_commit: bool = True) -> bool:
        """
        Add, updates or deletes a record related to a specific address and cube.
        To delete data, set argument ``data`` to ``None``.
        :param cube_name: Name of the targeted cube.
        :param records: A list of (address, data) tuples to be set for the specific cube.
        :param instant_commit: If set to ``True``, then the change is instantly committed to the underlying database.
        :return: ``True`` if successful.
        """
        # todo: add encryption

        table = self.DATA_TABLE_PREFIX + cube_name
        # separate records that should be deleted and upserted.
        del_records = [(address,) for address, data in records if not data]
        if del_records:
            sql = f"DELETE FROM {table} WHERE address = '?';"
            self._execute(sql, del_records)
            if self.encryptor:
                upsert_records = [(address, self.encryptor.encrypt(data)) for address, data in records if data]
            else:
                upsert_records = [(address, data) for address, data in records if data]
        else:
            upsert_records = records

        if upsert_records:
            sql = f"INSERT INTO {table}(address, data) VALUES(?,?) " \
                  f"ON CONFLICT(address) DO UPDATE SET data=EXCLUDED.data;"
            self._execute(sql, upsert_records)

        if instant_commit:
            self._commit()
        return True

    def get_record(self, cube_name: str, address: str):
        """
        Returns data from a cubes data table.
        :param cube_name: Name of the cube to get data from.
        :param address: The requested address.
        :return: The data stored for the given address.
           If the address does not exist, ``None`` will be returned.
        """
        # todo: add encryption

        table = self.DATA_TABLE_PREFIX + cube_name
        sql = f"SELECT data FROM {table} WHERE address = '{address}';"
        records = self._fetchall(sql)
        if records:
            if self.encryptor:
                return self.encryptor.decrypt(records[0][0])
            else:
                return records[0][0]
        return None

    def get_records(self, cube_name: str):
        """
        Returns all records from a cubes data table.
        :param cube_name: Name of the cube to get data from.
        :return: All records of the cube.
        """
        # todo: add encryption

        table = self.DATA_TABLE_PREFIX + cube_name
        sql = f"SELECT * FROM {table};"
        records = self._fetchall(sql)
        if records:
            return records
        return None

    # endregion

    # region meta data related methods
    def add_meta(self, key: str, json: str):
        """
        Adds or updates a meta configuration.
        :param key: The name of the meta item.
        :param json: The configuration of the meta item in json format.
        """
        if self.logging:
            self.logger.info(f"{self.LOG_PREFIX}Adding or updating meta configuration for key: '{key}'.")
            # self.logger.handlers[0].flush()

        key = self.encryptor.encrypt(key)
        json = self.encryptor.encrypt(json)

        data = (key, json)
        sql = f"INSERT INTO {self.META_TABLE_META} (key, config)" \
              f"VALUES(?, ?)  " \
              f"ON CONFLICT(key) DO UPDATE SET config = EXCLUDED.config;"
        self._execute(sql, data)
        self._commit()

    def get_meta(self, key: str) -> str:
        """
        Returns the configuration of a meta item in json format.
        :return: A json string.
        """
        key = self.encryptor.encrypt(key)

        result = self._fetchall(f"SELECT config FROM {self.META_TABLE_META} WHERE key = '{key}'")
        if result:
            value = str(result[0][0])
            value = self.encryptor.decrypt(value)
            return value
        return None

    # endregion

    # region Cube related methods
    def get_cubes(self) -> list[tuple[str, str]]:
        """
        Returns a list of all cubes and their configuration (in json format)
        available in the database.
        :return: A list of tuples of type (cube_name:str, json:str).
        """
        # todo: add encryption

        result = self._fetchall(f"SELECT * FROM {self.META_TABLE_CUBES}")
        return result

    def get_cube_names(self) -> list[str]:
        """
        Returns a list of all cubes names available in the database.
        :return: List of cube names.
        """
        # todo: add encryption

        names = []
        result = self._fetchall(f"SELECT key FROM {self.META_TABLE_CUBES}")
        for row in result:
            names.append(row[0])
        return names

    def count_cubes(self):
        """
        Returns the number of cubes defined in the database.
        :return: The number of cubes defined for the database.
        """
        return self._fetchall(f"SELECT COUNT(*) FROM {self.META_TABLE_CUBES}")[0][0]

    def add_cube(self, cube_name: str, json: str) -> bool:
        """
        Adds or updates a cube configuration.
        :param dimensions_count: Number of the dimension in the cube.
          If set to -1, then no data table will be created.
        :param cube_name: The name of the cube.
        :param json: The configuration of the cube in json format.
        :return: ``True`` if successful,``False`` otherwise.
        """
        # todo: add encryption

        if self.logging:
            self.logger.info(f"{self.LOG_PREFIX}Adding or updating configuration for cube '{cube_name}'.")
            # self.logger.handlers[0].flush()
        # add cube to meta table
        data = (cube_name, json)
        sql = f"INSERT INTO {self.META_TABLE_CUBES} (key, config)" \
              f"VALUES(?, ?)  " \
              f"ON CONFLICT(key) DO UPDATE SET config=EXCLUDED.config;"
        self._execute(sql, data)

        # create cube data table
        table_name = self.DATA_TABLE_PREFIX + cube_name
        sql = f"CREATE TABLE IF NOT EXISTS {table_name} " \
              f"(address TEXT PRIMARY KEY, data TEXT" \
              f") WITHOUT ROWID;"
        self._execute(sql)

        self._commit()
        return True

    def remove_cube(self, cube_name: str):
        """
        Removes a cube from the database.
        :param cube_name: Name of the cube to be removed.
        """
        # todo: add encryption

        if self.logging:
            self.logger.info(f"{self.LOG_PREFIX}Removing configuration and data tables for cube '{cube_name}'.")
            # self.logger.handlers[0].flush()
        table_name = self.DATA_TABLE_PREFIX + cube_name
        self._execute(f"DROP TABLE IF EXISTS {table_name};")
        self._execute(f"DELETE FROM {self.META_TABLE_CUBES} WHERE key = '{cube_name}';")
        self._commit()

    def clear_cube(self, cube_name: str) -> True:
        """
        Clears (deletes) all records from a cubes data table.
        :param cube_name: The name of the cube to be cleared.
        :return:
        """
        table_name = self.DATA_TABLE_PREFIX + cube_name
        if self._table_exists(table_name):
            self._execute(f"DELETE FROM {table_name};")
            self._commit()
        return True

    def count_cube_records(self, cube_name: str) -> int:
        """
        Returns the number of records contained in cube table.
        :return: The number of records in cube table.
        """
        # todo: add encryption

        table_name = self.DATA_TABLE_PREFIX + cube_name
        return self._fetchall(f"SELECT COUNT(*) FROM {table_name}")[0][0]

    # endregion

    # region Dimension related methods
    def get_dimensions(self) -> list[tuple[str, str]]:
        """
        Returns a list of all dimensions and their configuration (in json format)
        available in the database.
        :return: A list of tuples of type (cube_name:str, json:str).
        """
        # todo: add encryption

        result = self._fetchall(f"SELECT * FROM {self.META_TABLE_DIMENSIONS}")
        if result:
            return result
        return []

    def get_dimension_names(self) -> list[str]:
        """
        Returns a list of all dimension names available in the database.
        :return: List of database names.
        """
        # todo: add encryption

        names = []
        result = self._fetchall(f"SELECT key FROM {self.META_TABLE_DIMENSIONS}")
        for row in result:
            names.append(row[0])
        return names

    def count_dimensions(self):
        """
        Returns the number of dimensions defined in the database.
        :return: The number of dimension defined for the database.
        """
        # todo: add encryption

        return self._fetchall(f"SELECT COUNT(*) FROM {self.META_TABLE_DIMENSIONS}")[0][0]

    def add_dimension(self, dimension_name: str, json: str) -> bool:
        """
        Adds or updates a dimension configuration.
        :param dimension_name: The name of the dimension.
        :param json: The configuration of the dimension in json format.
        :return: ``True`` if successful,``False`` otherwise.
        """
        # todo: add encryption

        if self.logging:
            self.logger.info(f"{self.LOG_PREFIX}Adding or updating configuration for dimension '{dimension_name}'.")
            # self.logger.handlers[0].flush()
        data = (dimension_name, json)
        sql = f"INSERT INTO {self.META_TABLE_DIMENSIONS} (key, config)" \
              f"VALUES(?, ?)  " \
              f"ON CONFLICT(key) DO UPDATE SET config = EXCLUDED.config;"
        self._execute(sql, data)
        self._commit()

    def remove_dimension(self, dimension_name: str):
        """
        Removes a dimension from the database.
        :param dimension_name: Name of the dimension to be removed.
        """
        # todo: add encryption

        if self.logging:
            self.logger.info(f"{self.LOG_PREFIX}Removing configuration for dimension '{dimension_name}'.")
            # self.logger.handlers[0].flush()
        sql = f"DELETE FROM {self.META_TABLE_DIMENSIONS} WHERE key = '{dimension_name}';"
        self._execute(sql)
        self._commit()

    # endregion

    # region database related internal methods
    def _commit(self, optimize=False) -> True:
        """
        Commits all accumulated changes to the database. Optionally, the database can be optimized (compacted).
        :param optimize: if set to ``True``, the database will be optimized and compacted.
        :return: ``True`` if the commit and the optimization was successful.
        :raises DatabaseBackendException: Raised when the commit or (optional) optimization of the database failed.
        """
        if optimize:
            try:
                if self.logging:
                    self.logger.info(f"{self.LOG_PREFIX}Optimizing database (PRAGMA vacuum; PRAGMA optimize;).")
                    # self.logger.handlers[0].flush()
                self._execute('PRAGMA vacuum;')
                self._execute('PRAGMA optimize;')
            except sqlite3.Error as err:
                msg = f"{self.LOG_PREFIX}Failed to optimize database. SQLite exception: {str(err)}"
                if self.logging:
                    self.logger.error(msg)
                    # self.logger.handlers[0].flush()
                raise DatabaseBackendException(msg)
            except Exception as err:
                msg = f"{self.LOG_PREFIX}Failed to optimize database. {str(err)}"
                if self.logging:
                    self.logger.error(msg)
                    # self.logger.handlers[0].flush()
                raise DatabaseBackendException(msg)

        try:
            if self.conn and self.is_open:
                self.conn.commit()
            return True
        except sqlite3.Error as err:
            msg = f"{self.LOG_PREFIX}Failed to commit database. SQLite exception: {str(err)}"
            if self.logging:
                self.logger.error(msg)
                # self.logger.handlers[0].flush()
            raise DatabaseBackendException(msg)
        except Exception as err:
            msg = f"{self.LOG_PREFIX}Failed to commit database. {str(err)}"
            if self.logging:
                self.logger.error(msg)
                # self.logger.handlers[0].flush()
            raise DatabaseBackendException(msg)

    def _prepare_database(self) -> bool:
        """
        Validates and prepares a database to be used as a TinyOlap storage_provider
        by adding some meta tables.
        :return: ``True`` if the database was successfully prepared, ``False`` otherwise.
        :raises DatabaseBackendException: Raised when the preparation of the database failed.
        """

        # Check if meta tables already exist - then it's (most likely) a valid TinyOlap database.
        if not (self._table_exists(self.META_TABLE_CUBES) and self._table_exists(self.META_TABLE_DIMENSIONS)):
            # ...not yet a TinyOlap database, let's configure it.
            if self.logging:
                self.logger.info(f"{self.LOG_PREFIX}Prepare (new) database for initial use by adding meta tables.")
                # self.logger.handlers[0].flush()

            try:
                # do some configuration for optimized performance
                self._execute('PRAGMA temp_store=MEMORY;')
                self._execute('PRAGMA journal_mode=MEMORY;')
                self._commit()

                # add meta tables
                metas = [self.META_TABLE_META, self.META_TABLE_CUBES, self.META_TABLE_DIMENSIONS]
                for meta in metas:
                    if not self._table_exists(meta):
                        self._add_table(meta, self.META_TABLE_FIELDS)
                        self.add_meta("provider", "tinyolap")
                    if not self._table_exists(meta):  # ensure the table has been created.
                        if self.logging:
                            self.logger.info(f"{self.LOG_PREFIX}Failed to prepare database for use with TinyOlap.")
                        raise DatabaseBackendException("Failed to add meta tables to database.")

                # add and prepare history table
                if not self._table_exists(self.HISTORY_TABLE):
                    self._add_table(self.HISTORY_TABLE, self.HISTORY_TABLE_FIELDS)
                    self._execute(f"CREATE INDEX IF NOT EXISTS {self.HISTORY_TABLE}_timestamp_index "
                                  f"ON {self.HISTORY_TABLE}(timestamp);")
                    self._execute(f"CREATE INDEX IF NOT EXISTS {self.HISTORY_TABLE}_user_index "
                                  f"ON {self.HISTORY_TABLE}(user);")

            except sqlite3.Error as err:
                msg = f"{self.LOG_PREFIX}Failed to prepare database for use with TinyOlap. SQLite exception: {str(err)}"
                if self.logging:
                    self.logger.error(msg)
                    # self.logger.handlers[0].flush()
                raise DatabaseBackendException(msg)
            except Exception as err:
                msg = f"{self.LOG_PREFIX}Failed to prepare database for use with TinyOlap. {str(err)}"
                if self.logging:
                    self.logger.error(msg)
                    # self.logger.handlers[0].flush()
                raise DatabaseBackendException(msg)

        else:
            # check if data can be decrypted properly.
            provider = self.get_meta("provider")  # this should return 'tinyolap'
            if provider != "tinyolap":
                raise EncryptionException(f"Failed to open database. Invalid username or password.")

        return True

    def _add_table(self, table_name: str, fields: list[tuple[str, str]], index_sql: str = None):
        """
        Adds a new table to the database. If the table already exists,
        then the table will be dropped and recreated.
        :param table_name: Name of the table to be created.
        :param fields: A list of tuples containing field name and type
        """
        if self.logging:
            self.logger.info(f"{self.LOG_PREFIX}Adding or replacing table '{table_name}' to database.")
            # self.logger.handlers[0].flush()
        sql = f"DROP TABLE IF EXISTS {table_name};"
        self._execute(sql)
        sql = f"CREATE TABLE IF NOT EXISTS {table_name} ({', '.join([str(f[0]) + ' ' + str(f[1]) for f in fields])} );"
        self._execute(sql)
        self._commit()
        if index_sql:
            self._execute(index_sql)
            self._commit()
        return True

    def _table_exists(self, table_name) -> bool:
        """
        Evaluates if a table exists in the database.
        :param table_name: Name of the table to be evaluated.
        :return: ``True`` if the table exists, ``False`` otherwise.
        :raises DatabaseBackendException: Raised when the execution of the existence check failed.
        """
        if not self.cursor:
            self.cursor = self.conn.cursor()
        sql = f"SELECT count(name) FROM sqlite_master WHERE type='table' AND name='{table_name}'"
        try:
            return self.cursor.execute(sql).fetchone()[0] == 1
        except (sqlite3.Error, Exception) as err:
            msg = f"{self.LOG_PREFIX}Failed to check existence of table {table_name} in database. {str(err)}"
            if self.logging:
                self.logger.error(msg)
            raise DatabaseBackendException(msg)

    def _get_tables(self):
        """
        Returns the list of all tables available in the database.
        :return: A resultset containing the tables of the database.
        """
        return self._fetchall(f"SELECT name, sql FROM sqlite_master WHERE type='table';")

    def _execute(self, sql: str, data=None) -> bool:
        """
        Executes an SQL statement without returning a result or resultset.
        Intended to be used for DDL statements.
        :param sql: The SQL statement to be executed.
        :param data: (optional) data to be handed in.
        :return: ``True`` is successful, ``False`` otherwise.
        """
        duration = 0.0
        if self.logging and self.LOG_LEVEL == logging.DEBUG:
            duration = timer()

        try:
            if not self.cursor:
                self.cursor = self.conn.cursor()
            if data:
                if type(data) is list:
                    result = self.cursor.executemany(sql, data)
                else:
                    result = self.cursor.execute(sql, data)
            else:
                result = self.cursor.execute(sql)

        except (sqlite3.Error, Exception) as err:
            msg = f"{self.LOG_PREFIX}Failed to execute SQL statement. {str(err)} SQL := {sql}"
            if self.logging:
                self.logger.error(msg)
                # self.logger.handlers[0].flush()
            raise DatabaseBackendException(msg)

        if self.logging and self.LOG_LEVEL == logging.DEBUG:
            duration = timer() - duration
            self.logger.debug(f"{self.LOG_PREFIX}SQL statement successfully executed in {duration:.6f}s: {sql}")
        return True

    def _execute_transaction(self, sql: list[str]) -> bool:
        """
        Executes a list of SQL statements within a single transaction (rollback on failure).
        :param sql: A list of SQL statement to be executed.
        :return: ``True`` is successful, ``False`` otherwise.
        """
        duration = 0.0
        if self.logging and self.LOG_LEVEL == logging.DEBUG:
            duration = timer()

        current_statement = ""
        if not self.cursor:
            self.cursor = self.conn.cursor()
        try:
            # clear current dim table
            self.cursor.execute("begin")
            for statement in sql:
                current_statement = statement
                self.cursor.execute(statement)
            self.cursor.execute("commit")
        except (sqlite3.Error, Exception) as err:
            msg = f"{self.LOG_PREFIX}Failed to execute transaction containing {len(sql)} statements." \
                  f" {str(err)} Failed SQL := {current_statement}"
            if self.logging:
                self.logger.error(msg)
                # self.logger.handlers[0].flush()
            self.cursor.execute("rollback")
            raise DatabaseBackendException(msg)

        if self.logging and self.LOG_LEVEL == logging.DEBUG:
            duration = timer() - duration
            self.logger.debug(f"{self.LOG_PREFIX}SQL transaction executed in {duration:.6f}s: {sql}")
        return True

    def _fetchall(self, sql: str):
        """
        Executes an SQL statement and returns all records as the resultset.
        :param sql: The SQL statement to be executed.
        :return: The resultset of the query.
        """
        duration = 0.0
        if self.LOG_LEVEL == logging.DEBUG:
            duration = timer()

        try:
            if not self.cursor:
                self.cursor = self.conn.cursor()
            result = self.cursor.execute(sql).fetchall()
        except (sqlite3.Error, Exception) as err:
            msg = f"{self.LOG_PREFIX}Failed to execute SQL statement and fecth results. {str(err)} SQL := {sql}"
            if self.logging:
                self.logger.error(msg)
                # self.logger.handlers[0].flush()
            raise DatabaseBackendException(msg)

        if self.logging and self.LOG_LEVEL == logging.DEBUG:
            duration = timer() - duration
            self.logger.error(f"{self.LOG_PREFIX}SQL fetchall executed in {duration:.6f}s: {sql}")

        return result

    # endregion

    # region file handling & naming
    def _evaluate_path(self, name: str) -> tuple[bool, str, Path, str]:
        """
        Tries to evaluate a valid file path from a given name or file path.
        :param name: The name or file path to be evaluated.
        :return: A tuple of type (exists:bool, folder:str, path_to_file:str, file_name:str).
          The *exists* flag identifies if a file for the given name already exists.
        """
        # check if the name is a file path and maybe already exists
        file = Path(name)
        exists = file.exists() & file.is_file()
        if exists:
            file_path = file.absolute()
            file_name = file.name
            folder = file.parent.absolute()
            return bool(exists), str(folder), file_path, file_name

        # ...file does not exist, setup a valid file apth from the predefined or default file path.
        file_name = name
        if not file_name.endswith(self.DB_EXTENSION):
            file_name += self.DB_EXTENSION

        if not self.database_folder:  # use default database location
            folder = os.path.join(os.getcwd(), self.DB_DEFAULT_FOLDER_NAME)
        else:
            folder = self.database_folder

        # Ensure the database folder exists, if not create it.
        try:
            if not os.path.exists(folder):
                os.makedirs(folder, exist_ok=True)
                if self.logging:
                    self.logger.error(f"Database folder '{folder}' has been created.")
                    # self.logger.handlers[0].flush()
        except OSError as err:
            msg = f"Failed to create database folder '{folder}'. {str(err)}"
            if self.logging:
                self.logger.error(msg)
                # self.logger.handlers[0].flush()
            raise DatabaseBackendException(msg)

        # Assemble database file name
        file_path = os.path.join(folder, file_name)
        file = Path(file_path)
        exists = file.exists() & file.is_file()
        file_path = file.absolute()
        folder = file.parent.absolute()
        file_name = file.name
        return bool(exists), str(folder), file_path, file_name

    def _to_save_name(self, name: str) -> str:
        """
        Converts a string to a valid filename.
        :param name: The name to be converted.
        :return: A valid file_name, without special characters.
        """
        valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
        valid_name = ''.join(c for c in name if c in valid_chars)
        return valid_name.replace(' ', '_')

    def _initialize_logger(self):
        """Initializes the logger, if required."""
        if self.logger:  # an existing logger was already handed in
            return
        if not self.logging:  # enable_logging not required.
            return

        self.logger = logging.getLogger("tinyolap.storage_provider.sqlite")

        if not self.log_file:
            file_exists, self.folder, self.file_path, self.file_name = self._evaluate_path(self.name)
            self.log_file = str(self.file_path) + self.LOG_EXTENSION

        handler = logging.FileHandler(self.log_file, mode='w')
        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(self.LOG_LEVEL)

        # self.logger = enable_logging.getLogger("TinOlap.SqliteStorage")
        # handler = enable_logging.FileHandler(self.log_file, mode='w')
        # handler.setLevel(self.LOG_LEVEL)
        # formatter = enable_logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        # handler.setFormatter(formatter)
        # self.logger.addHandler(handler)

    # endregion



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations
from enum import Enum, IntEnum, IntFlag
import inspect
import enum_tools.documentation

enum_tools.documentation.INTERACTIVE = True


@enum_tools.documentation.document_enum
class RuleError(Enum):
    """Defines error values raised by rules, useing an Excel-alike error format."""
    DIV0 = "#DIV/0!"  # doc: Devision by zero.
    VALUE = "#VALUE!"  # doc: Invalid argument type, e.g. a number was used instead of an expected string.
    REF = "#REF!"  # doc: Invalid reference to a member, dimension, attribute or cube name.
    ERROR = "#ERR!"  # doc: An unknown or unhandable error occurred.


@enum_tools.documentation.document_enum
class RuleScope(IntEnum):
    """
    Defines the scope of a rule. Meaning, to which level of data the rule should be applied.
    """
    ALL_LEVELS = 1  # doc: (default) Indicates that the rule should be executed for base level and aggregated level cells.
    AGGREGATION_LEVEL = 2  # doc: Indicates that the rule should be executed for aggregated level cells only.
    BASE_LEVEL = 3  # doc: Indicates that the rule should be executed for base level cells only.
    ROLL_UP = 4  # doc: Indicates that the rule should replace the base level cell value from the database by the results of the rule. This can dramatically slow down aggregation speed. Requires a special trigger to be set.
    ON_ENTRY = 5  # doc: Indicates that these rules should be executed when cell values are set or changed. This is useful for time consuming calculations which may be *too expensive* to run at idx_address time.
    COMMAND = 6  # doc: Indicates that these rules need to be invoked by a command. Requires the decorator parameter 'command to be specified.

    def __eq__(self, other):
        return self.value == int(other)

    def __ne__(self, other):
        return self.value != other.value

    def __hash__(self):
        return hash(self.value)


@enum_tools.documentation.document_enum
class RuleInjectionStrategy(IntEnum):
    """
    Defines the code injection strategy for individual rules.

    By default, TinyOlap rules reside and will be executed from with your custom code.
    This is preferable for a lot of situations, e.g. for development and debugging,
    or when your business logic require resources that can or should not become a part
    of a TinyOlap database, like calling other systems or systems. When a TinyOlap
    database is running in in-memory mode this is anyhow the only available option
    to provide business logic to a TinyOlap database.

    However, when you intend to hand over a TinyOlap database to someone else, or if
    you want to host it as a generic webservice, then your business logic ideally
    goes with the database.

    To enable this TinyOlap can automatically inject your rule source code into the
    database and persist it with the database. The next time the database will be opened,
    your code will be automatically instantiated and run from within the TinyOlap engine
    itself. You can at anytime override / replace these injected rules with your own
    code by calling  the ``add_rule(...)`` method provide by the *cube* class.

    There are 4 different strategies available how to inject rules into a TinyOlap
    database. Depending on your use case, you should try to use the most restrictive
    strategy possible, as explained below in the documentation of the different
    RuleInjectionStrategy enum values.

    Please be aware that code injection does work on actual code level, not on file level.
    If you have created dynamic code, the code should be properly extracted by TinyOlap.
    """

    NO_INJECTION = 0  # doc: (default) Indicates that the rule should not be injected into the database.
    FUNCTION_INJECTION = 1  # doc: Indicates that **only** the rule function itself will be injected into the database. All surrounding code of the module or project where the rule function is defined, will be ignored. This requires your rule function to be **autonomous**. Meaning, without any dependencies to functions or classes from within your code. By default, TinyOlap will only reference the following built-in Python modules using the ``from [module name] import *`` trigger, when running your code: math, cmath, statistics, decimal, fractions, random, datetime, time, re, json. ``FUNCTION_INJECTION`` should be the preferred strategy for simple business logic that acts upon the data from a TinyOlap database only.
    MODULE_INJECTION = 2  # doc: Indicates that the **entire** module in which the rule function is defined will be injected doc: into the database. Code from other modules of your project will be ignored. If your rules are spread or multiple modules, all these modules will be injected. This requires that all modules and Python packages referenced from within your module must also be installed on the target system. TinyOlap will raise an appropriate error if the instantiation of your code module in the target environment will fail. ``MODULE_INJECTION`` should be the preferred strategy for more complex business logic or business logic that requires certain initialization (e.g. read exchange rates from a service)
    PROJECT_INJECTION = 3  # doc: **NOT YET SUPPORTED** Indicates that the **entire** project in which the rule function is defined will be injected into the database. This

    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    def __hash__(self):
        return hash(self.value)


class Rule:
    """
    Represents a rule, defining custom calculations or business logic to be assigned to a cube.
    """

    def __init__(self, function, name: str, cube: str, trigger: list[str], idx_trigger_pattern: list[tuple[int, int]],
                 scope: RuleScope, injection: RuleInjectionStrategy, code: str = None):
        self.function = function
        self.cube: str = cube
        self.name: str = name
        self.trigger: list[str] = trigger
        self.idx_trigger_pattern = idx_trigger_pattern
        self.scope: RuleScope = scope
        self.injection: RuleInjectionStrategy = injection
        self.code: str = code
        self.signature: str = self.cube + str(idx_trigger_pattern)


class Rules:
    """Represemts a list of rules. Rules define custom calculations or business logic to be assigned to a cube.

    Rules consist two main components:

    * A trigger or trigger, defining the context for which the rule should be executed

    * A scope, defining to which level of data the rule should be applied.
      Either for base level cells, aggregated cells, all cells or on write back of values.

    * A function, defining the custom calculation or business logic. This can be any Python method or function.

    .. attention::

        Rules functions have to be implemented as a simple Python function with just one single parameter and
        a return value. The single parameter should be called 'c' and will contain an TinyOlap Cell, representing
        the current cell context the rule should be calculated for.

        What happens in a rule function, is up totally to the programmer. The value returned by rules function
        can either be a certain value (most often a numerical number, but can be anything) or one of the following
        constants which are directly available from within a cursor object.

        * **NONE** - Indicates that rules function was not able return a proper result (why ever).

        * **CONTINUE** - Indicates that either subsequent rules should continue and do the calculation work
           or that the cell value, either from a base-level or an aggregated cell, form the underlying cube should
           be used.

        * **ERROR** - Indicates that the rules functions run into an error. Such errors will be pushed up to initially
          calling cell request.

        Sample of a proper rule:

        .. code:: python

            def rule_average_price(c : tinyolap.context):
                quantity = c["quantity"]
                sales = c["sales"]
                # ensure both values exist or are of the expected type (cell values can be anything)
                if quantity is float and sales is float:
                    if quantity != 0.0:
                        return sales / quantity
                    return "n.a."  # the developer decided to return some text, what is totally fine.
                return c.CONTINUE
    """

    def __init__(self, cube: str):
        # new implementation
        self.rules: dict[RuleScope, list[Rule]] = {}
        self.patterns: dict[RuleScope, list[list[tuple[int, int]]]] = {}

        # old implementation
        self.any: bool = False
        self.cube = cube
        self.functions = []
        self.function_names = []
        self.function_scopes = []
        self.function_injections = []
        self.source = []
        self.pattern = []
        self.pattern_idx = []

    def add(self, rule: Rule):
        """
        Adds a new rule to the list of rules. If a rule with the same trigger pattern
        already exists, then the existing rule will be replaceed by the new rule.
        :param rule: The Rule to be added.
        """
        scope = rule.scope
        idx_existing = -1
        if scope in self.rules:
            # Note: sorting matters, as we don't know the sequence in which users define rules !!!
            rule.idx_trigger_pattern.sort()
            for idx, existing_rule in enumerate(self.rules[scope]):
                # check if trigger pattern are identical
                if rule.idx_trigger_pattern == existing_rule.idx_trigger_pattern:
                    # replace existing rule with new rule.
                    idx_existing = idx
                    break
            if idx_existing >= 0:
                # replace rule (the trigger pattern is the same)
                self.rules[scope][idx_existing] = rule
            else:
                # add new rule
                self.rules[scope].append(rule)
                self.patterns[scope].append(rule.idx_trigger_pattern)
        else:
            # add first rule and pattern for this scope
            self.rules[scope] = [rule]
            self.patterns[scope] = [rule.idx_trigger_pattern]

    def match(self, scope: RuleScope, idx_address: list[tuple[int, int]]) -> (bool, object):
        """
        Returns the first trigger match, if any, for a given cell address.

        :param scope: The rule scope for which a matching rule is requested.
        :param idx_address: The cell address (in index int format) to be evaluated.
        :return: Returns a tuple (True, *function*) if at least one trigger matches,
            *function* is the actual rules function to call, or (False, None) if none
            of the patterns matches the given cell idx_address.

        """
        patterns = self.patterns.get(scope)
        if patterns:
            for idx, function_pattern in enumerate(patterns):  # e.g. [(0,3),(3,2)] >> dim0 = member3, dim3 = member2
                for dim_pattern in function_pattern:  # e.g. (0,3) >> dim0 = member3
                    if idx_address[dim_pattern[0]] != dim_pattern[1]:
                        break
                else:
                    return True, self.rules[scope][idx].function  # this statement will be executed only,
                    # if the inner loop did NOT break

        return False, None

    def __bool__(self):
        return self.functions is True

    def __len__(self):
        return len(self.functions)

    def register(self, function, cube: str, function_name: str,
                 pattern: list[str], idx_pattern: list[tuple[int, int]],
                 scope: RuleScope, injection: RuleInjectionStrategy, code: str = None):
        """
        Registers a rules function (a Python method or function).

        :param injection: The injection strategy defined for the rule.
        :param code: (optional) the source code of the rule.
        :param scope: The scope of the rule function.
        :param function_name: Name of the rule function.
        :param function: The Python rule function to execute.
        :param pattern: The cell trigger to trigger the rule function.
        :param idx_pattern: The cell index trigger to trigger the rule function.
        """
        self.functions.append(function)
        self.function_names.append(function_name)
        self.function_scopes.append(scope)
        self.function_injections.append(injection)
        self.pattern.append(pattern)
        if code:
            self.source.append(code)
        else:
            self.source.append(self._get_source(function))
        self.pattern_idx.append(idx_pattern)
        self.any = True

    # def _get_source_code(self, function):
    #     source = inspect.getsource(function)
    #     module = inspect.getmodule(function)
    #     module_source = inspect.getsource(module)
    #     sourcefile = inspect.getsourcefile(function)
    #     print(f"Rule from module {module} and file '{sourcefile}'.")
    #     print(module_source)
    #
    #     return source

    @staticmethod
    def _get_source(function):
        try:
            lines = inspect.getsource(function)
        except Exception as err:
            lines = None
        return lines

    def first_match(self, idx_address) -> (bool, object):
        """
        Returns the first trigger match, if any, for a given cell address.

        :param idx_address: The cell address in index number_format.
        :return: Returns a tuple (True, *function*) if at least one trigger matches,
            *function* is the actual rules function to call, or (False, None) if none
            of the patterns matches the given cell idx_address.

        """
        for idx, function_pattern in enumerate(
                self.pattern_idx):  # e.g. [(0,3),(3,2)] >> dim0 = member3, dim3 = member2
            for dim_pattern in function_pattern:  # e.g. (0,3) >> dim0 = member3
                if idx_address[dim_pattern[0]] != dim_pattern[1]:
                    break
            else:
                return True, self.functions[idx]  # this will be executed only if the inner loop did NOT break
        return False, None



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import itertools
from inspect import signature

import copy
from collections import Iterable

from tinyolap.member import Member


class Area:
    """
    An Area represents a subspace of a cube. Areas are intended be used for mass operations on cube cells.
    The following example should explain the basic concept:

    .. code:: python

        area = cube.area("Jan", "2023", "Plan")
        # clears all data for months='Jan', years='2023' and datatype='Plan' over all other dimensions.
        area.clear()

        # Copies all data from 'Actual', '2022', increases all values by 15% and writes them to the 'Plan' for '2023'.
        # Info: All data in the target area will be deleted before hand.
        # Info: Depending on the amount of data to be copied, such operations may take some time (a few seconds).
        cube.area("Plan", "2023") = cube.area("Actual", "2022") * 1,15

    The smallest possible subspace would be a single cube cell.
    For such purposes it is recommended to use the Cell object.
    """

    def __init__(self, cube, args):
        """
        Initializes a new data area for a specific cube.
        :param cube: The cube to create the data area for.
        :param args: The member specification for the data area.
        """
        self._cube = cube
        self._args = args
        self._area_def = []
        self._idx_area_def = []
        self._levels_area_def = []
        self._validate(args)
        self._rows = set()
        self._modifiers = []
        self._func = None
        self._pinned = False
        self._pinned_records = []

    def __len__(self):
        if not self._rows:
            # refresh dat area
            rows = self._cube._facts.query_area(self._idx_area_def)
            self._rows = rows

        return len(self._rows)

    # region Area manipulation via indexing/slicing
    def __getitem__(self, args):
        area = self.clone().alter(args)
        # We need to pin (freeze) the current state and data of the area
        # as source and target area in area-operations can overlap (or even be identical)
        # and would cause inconsistent operations. Lazy operations (using 'yield') would fail.
        area._pinned = True
        area._pinned_records = area._raw_records()
        return area

    def __setitem__(self, args, value):
        if type(value) is Area:
            area = self.clone().alter(args)
            area._apply_other_area(args, value)
        else:
            area = self.clone().alter(args)
            area.set_value(value)

    def __delitem__(self, args):
        item_area = self.clone().alter(args)
        item_area.clear()

    # endregion

    def records(self, include_cube_name: bool = False, as_list: bool = True):
        """
        Generator to loop over existing items of the data area.
        Returns nested tuples in the form ((dim_member1, ... , dim_memberN), value).
        """
        # if not self._rows:
        # refresh dat area
        rows = self._cube._facts.query_area(self._idx_area_def)
        self._rows = rows

        facts = self._cube._facts
        for row in self._rows:
            idx_address = facts.addresses[row]
            record = self._cube._idx_address_to_address(idx_address, include_cube_name=include_cube_name)

            if as_list:
                for key in facts.facts[row]:
                    value = facts.facts[row][key]
                    record.append(value)
            else:
                values = tuple(facts.facts[row].values())
                record = [tuple(record), values]

            yield record

    def _raw_records(self) -> list:
        """
        Generator to loop over existing items of the data area.
        Returns nested tuples in the form ((dim_member1, ... , dim_memberN), value).
        """
        # if not self._rows:
        # refresh dat area
        rows = self._cube._facts.query_area(self._idx_area_def)
        self._rows = rows
        records = []
        facts = self._cube._facts
        idx_measure = 0
        super_level = 0
        for row in self._rows:
            # idx_address = facts.addresses[row]
            record = list(facts.addresses[row])
            # for key in facts.facts[row]:
            #     value = facts.facts[row][key]
            #     record.append(value)

            # create a (idx_address, value) tuple for direct use
            records.append((record, dict(facts.facts[row])))
        return records

    def addresses(self, include_cube_name: bool = False, enumerate_data_space: bool = False):
        """
        Generator to loop over existing addresses of the data area.
        Returns tuples in the form (dim_member1, ... , dim_memberN).
        :param include_cube_name: Identifies if the name of the cube should be
        included in the address tuple.
        :param enumerate_data_space: Identifies if the entire data space, independent of
        value exists or not should be returned. Caution, the data space can be hugh.
        """
        if enumerate_data_space:
            member_lists =[]
            for dim_idx, axis in enumerate(self._idx_area_def):
                dimension = self._cube._dimensions[dim_idx]
                if axis:
                    member_list = []
                    if type(axis) is int:
                        member = dimension.members[axis][1]
                        member_list.extend(dimension.member_get_leave_children(member))
                    else:
                        for item in axis:
                            member = dimension.members[item][1]
                            member_list.extend(dimension.member_get_leave_children(member))
                    member_lists.append(member_list)
                else:
                    # get all base members from dimension
                    member_lists.append(dimension.get_leave_members())

            if include_cube_name:
                cube = [self._cube.name, ]
                for address in itertools.product(*member_lists):
                    yield cube + address
            else:
                for address in itertools.product(*member_lists):
                    yield address

        else:
            if not self._rows:
                # refresh dat area
                rows = self._cube._facts.query_area(self._idx_area_def)
                self._rows = rows

            facts = self._cube._facts
            for row in self._rows:
                idx_address = facts.addresses[row]
                yield self._cube._idx_address_to_address(idx_address, include_cube_name=include_cube_name)

    def enumerate(self, include_cube_name: bool = False):
        """
        Generator to loop over all addresses of the data area. It enumerates the entire space.
        Returns tuples in the form (dim_member1, ... , dim_memberN).
        """
        if not self._rows:
            # refresh dat area
            rows = self._cube._facts.query_area(self._idx_area_def)
            self._rows = rows

        facts = self._cube._facts
        for row in self._rows:
            idx_address = facts.addresses[row]
            yield self._cube._idx_address_to_address(idx_address, include_cube_name=include_cube_name)


    def alter(self, *args) -> Area:
        """Alters the data area, based on the given arguments."""
        self._validate(args, alter=True)
        # self._modifiers = []
        self.refresh()
        return self

    def clone(self) -> Area:
        """Creates a copy of the data area."""
        cloned = copy.copy(self)
        return cloned

    def refresh(self):
        self._rows = self._cube._facts.query_area(self._idx_area_def)

    def clear(self):
        """
        Clears the data area. All cells holding values will be removed from the cube.
        """
        if not self._rows:
            rows = self._cube._facts.query_area(self._idx_area_def)
        else:
            rows = self._rows
        self._cube._facts.remove_records(rows)
        self._rows = set()

    def set_value(self, value, enumerate_data_space: bool = False):
        """
        Sets all existing cells of the data area to a specific value.
        :param value: The value to be set.
        """
        if enumerate_data_space:
            rows = None
        else:
            rows = self._cube._facts.query_area(self._idx_area_def)
        if not rows:
            # empty data area
            # write to the entire data space
            if callable(value):
                for address in self.addresses(False, True):
                    self._cube.set(address, value())
            else:
                for address in self.addresses(False, True):
                    self._cube.set(address, value)
        else:
            self._rows = rows
            facts = self._cube._facts.facts
            if callable(value):
                for row in rows:
                    for k in facts[row]:
                        facts[row][k] = value()
            else:
                for row in rows:
                    for k in facts[row]:
                        facts[row][k] = value


    def multiply(self, factor: float):
        """
        Multiplies all existing cells holding numeric values with a certain factor.
        :type factor: The factor to multiply all cells holding numeric values with.
        """
        rows = self._cube._facts.query_area(self._idx_area_def)
        self._rows = rows
        facts = self._cube._facts.facts
        for row in rows:
            for k, v in facts[row].items():
                if type(v) is float:
                    facts[row][k] = v * factor

    def increment(self, value: float):
        """
        Increments all existing cells holding numeric values by a certain value.
        :type value: The value to increment all cells holding numeric values by.
        """
        rows = self._cube._facts.query_area(self._idx_area_def)
        self._rows = rows
        facts = self._cube._facts.facts
        for row in rows:
            for k, v in facts[row].items():
                if type(v) is float:
                    facts[row][k] = v + value

    def min(self):
        """
        Returns the minimum value of all existing numeric cells in the data area.
        """
        minimum = None
        rows = self._cube._facts.query_area(self._idx_area_def)
        self._rows = rows
        facts = self._cube._facts.facts
        for row in rows:
            for v in facts[row].values():
                if type(v) is float:
                    if not minimum:
                        minimum = v
                    if v < minimum:
                        minimum = v
        return minimum

    def max(self):
        """
        Returns the maximum value of all existing numeric cells in the data area.
        """
        maximum = None
        rows = self._cube._facts.query_area(self._idx_area_def)
        self._rows = rows
        facts = self._cube._facts.facts
        for row in rows:
            for v in facts[row].values():
                if type(v) is float:
                    if not maximum:
                        maximum = v
                    if v > maximum:
                        maximum = v
        return maximum

    def avg(self):
        """
        Returns the average of all existing numeric cells in the data area.
        """
        avg = 0.0
        z = 0
        rows = self._cube._facts.query_area(self._idx_area_def)
        self._rows = rows
        facts = self._cube._facts.facts
        for row in rows:
            for v in facts[row].values():
                if type(v) is float:
                    avg += v
                    z += 1
        if z == 0:
            return None
        return avg / z

    def sum(self):
        """
        Returns the sum of all existing numeric cells in the data area.
        """
        total = 0.0
        z = 0
        rows = self._cube._facts.query_area(self._idx_area_def)
        self._rows = rows
        facts = self._cube._facts.facts
        for row in rows:
            for v in facts[row].values():
                if type(v) is float:
                    total += v
                    z += 1
        if z == 0:
            return None
        return total

    def to_json(self, compact: bool = True):
        """
        Returns a data area as a json string, including the definition of the area
        and all base level cell data from that area. For larger data areas it is
        recommended to set parameter ``compact`` to ``True`` (the default value).
        This reduces the size of the returned json string by factor 2 to 10.

        The json returned by this method will identical to calling ``json.dumps(area.to_dict())`` .

        This function is intended to be used for custom persistence.
        If you just want to temporarily keep data (e.g. to undo certain changes),
        in memory, then consider to use method ``to_dict()`` for better performance
        instead.

        :param compact: Identifies is the data area should be compacted.
        """
        return NotImplemented

    def from_json(self, json: str, validate: bool = False) -> bool:
        """
        Reads a data area from a json string, clears the data area and
        writes (imports) all values contained in the data area definition into the cube.

        :param json: A json string containing the data area definition.
        :param validate: If set to ``True`` , the data area will be validated before importing.
        By this you can ensure that all values of the data area can be written to the cube.
        If at least one value of the data area can not be written to the cube, then ```from_json()``
        will stop and return ``False`` , otherwise the method will continue and start to write
        (import) values to the cube.
        :return Returns ``True``if the import was successful, ``False``otherwise.
        """
        return NotImplemented

    def to_dict(self, compact: bool = True):
        """
        Returns a data area as a Python ``dict`` object. For larger data areas it is
        recommended to set parameter ``compact`` to ``True`` (the default value).
        This reduces the memory footprint of the returned dict by factor 2 to 10.

        :param compact: Identifies is the data area should be compacted.
        """
        return NotImplemented

    def from_dict(self, area_dict: dict, validate: bool = False) -> bool:
        """
        Reads a data area from a Python dict object, clears the data area and
        writes (imports) all values contained in the data area definition into the cube.

        :param area_dict: A Python dict object containing the data area definition.
        :param validate: If set to ``True`` , the data area will be validated before importing.
        By this you can ensure that all values of the data area can be written to the cube.
        If at least one value of the data area can not be written to the cube, then ```from_dict()``
        will stop and return ``False`` , otherwise the method will continue and start to write
        (import) values to the cube.
        :return Returns ``True``if the import was successful, ``False``otherwise.
        """
        return NotImplemented

    # region operator overloading for numerical operations
    def __add__(self, other):  # + operator
        self._func = lambda x: x + other
        return self

    def __iadd__(self, other):  # += operator
        self._func = lambda x: x + other
        return self

    def __radd__(self, other):  # + operator
        self._func = lambda x: x + other
        return self

    def __sub__(self, other):  # - operator
        self._func = lambda x: x - other
        return self

    def __isub__(self, other):  # -= operator
        self._func = lambda x: x - other
        return self

    def __rsub__(self, other):  # - operator
        self._func = lambda x: x - other
        return self

    def __mul__(self, other):  # * operator
        self._func = lambda x: x * other
        return self

    def __imul__(self, other):  # *= operator
        self._func = lambda x: x * other
        return self

    def __rmul__(self, other):  # * operator
        self._func = lambda x: x * other
        return self

    def __truediv__(self, other):  # / operator (returns an float)
        self._func = lambda x: x / other
        return self

    def __idiv__(self, other):  # /= operator (returns an float)
        self._func = lambda x: x / other
        return self

    def __rtruediv__(self, other):  # / operator (returns an float)
        self._func = lambda x: x / other
        return self

    def __floordiv__(self, other):  # // operator (returns an integer)
        self._func = lambda x: x // other
        return self

    def __ifloordiv__(self, other):  # //= operator (returns an integer)
        self._func = lambda x: x // other
        return self

    def __rfloordiv__(self, other):  # // operator (returns an integer)
        self._func = lambda x: x // other
        return self

    # endregion

    def _validate(self, args, alter: bool = False):

        # reset area definition
        dim_count = self._cube.dimensions_count

        if alter:
            self._modifiers = []
        else:
            self._area_def = [None] * dim_count
            self._idx_area_def = [None] * dim_count
            self._levels_area_def = [None] * dim_count

        already_used = set()
        for arg in args:
            idx_dim, members, idx_members, level_members = self.__get_members(arg)

            if not alter:
                if idx_dim in already_used:
                    raise TypeError(f"Duplicate member definition argument '{str(arg)}'. The dimension of these members"
                                    f"have already been defined in the area definition.")
                already_used.add(idx_dim)
            else:
                self._modifiers.append([idx_dim, members, idx_members, level_members])

            self._area_def[idx_dim] = members
            self._idx_area_def[idx_dim] = idx_members
            self._levels_area_def[idx_dim] = level_members

    def __get_members(self, item):
        idx_dims = []
        members = []
        idx_members = []
        level_members = []
        dim_names = [key for key in self._cube._dim_lookup.keys()]

        if type(item) is str or not isinstance(item, Iterable):
            item = (item,)

        for member in item:
            if type(member) is Member:
                idx_dims.append(member._idx_dim)
                members.append(member.name)
                idx_members.append(member._idx_member)
                level_members.append(member._member_level)

            elif type(member) is str:
                idx_dim, idx_member, member_level = self._get_member(dim_names, member)
                idx_dims.append(idx_dim)
                members.append(member)
                idx_members.append(idx_member)
                level_members.append(member_level)
            else:
                raise TypeError(f"Invalid type '{type(item)}'. Only type 'str', 'list' or 'tuple' "
                                f"supported for members in Area definition.")

        # ensure all members are from the same dimension
        # todo: NO!!! Wrong behaviour! Areas shifts should support multiple dimensions.
        #       We need this to be possible: ... area(('Jan', 'Feb), '2021', 'name of subset')
        idx_dim = idx_dims[0]
        for idx in idx_dims[1:]:
            if idx != idx_dim:
                raise TypeError(f"Invalid member definition argument '{str(item)}'. Members do not belong "
                                f"to one dimension only, multiple dimensions found.")

        return idx_dim, members, idx_members, level_members

    def _get_member(self, dim_names, member_name: str):
        level = self._cube._dimensions[0].LEVEL
        dimensions = self._cube._dimensions
        idx_dim = -1
        idx_member = -1
        pos = member_name.find(":")
        if pos != -1:
            # lets extract the dimension name and check if it is valid, e.g., c["months:Mar"]
            dim_name = member_name[:pos].strip()

            # special test for ordinal dim position instead of dim name, e.g., c["1:Mar"] = 333.0
            if dim_name.isdigit():
                ordinal = int(dim_name)
                if 0 <= ordinal < len(dim_names):
                    # that's a valid dimension position number
                    idx_dim = ordinal
            if idx_dim == -1:
                if dim_name not in self._cube._dim_lookup:
                    raise KeyError(f"Invalid member key. '{dim_name}' is not a dimension "
                                   f"in cube '{self._cube.name}. Found in '{member_name}'.")
                idx_dim = self._cube._dim_lookup[dim_name]

            # adjust the member name
            member_name = member_name[pos + 1:].strip()
            if member_name not in dimensions[idx_dim]._member_idx_lookup:
                raise KeyError(f"Invalid member key. '{member_name}'is not a member of "
                               f"dimension '{dim_name}' in cube '{self._cube.name}.")
            idx_member = dimensions[idx_dim]._member_idx_lookup[member_name]

            member_level = dimensions[idx_dim].members[idx_member][self._cube._dimensions[0].LEVEL]
            return idx_dim, idx_member, member_level

        # No dimension identifier in member name, search all dimensions
        for idx_dim in range(self._cube.dimensions_count):
            if member_name in dimensions[idx_dim]._member_idx_lookup:
                idx_member = dimensions[idx_dim]._member_idx_lookup[member_name]
                # adjust the super_level
                member_level = dimensions[idx_dim].members[idx_member][level]
                return idx_dim, idx_member, member_level

        # You loose...
        if idx_dim == -1:
            raise KeyError(f"'{member_name}' is not a member of any dimension in "
                           f"cube '{self._cube.name}', or a valid reference to any of it's dimensions.")

    def _apply_other_area(self, args, other: Area):
        if not (self._cube.name is other._cube.name):
            raise KeyError(f"Unsupported area operations. Areas need to be from one cube, "
                           f"but one area is from cube '{self._cube.name}' "
                           f"and the other from cube '{other._cube.name}' .")

        if self._modifiers:

            if not self.modifiers_of_same_scope(self._modifiers, other._modifiers):
                raise KeyError(f"Unsupported area operations. Areas modifiers need to"
                               f"address the same dimensions and the same number of members, "
                               f"e.g.: a['Jan'] = a['Feb'] would be valid, but "
                               f"a['Jan'] = a['2022'] would not be valid, because of different dimensions.")

            if not self.identical_modifiers(self._modifiers, other._modifiers):
                self.clear()  # clear the target area

            if other._pinned:
                for record in other._pinned_records:
                    idx_address = record[0]
                    values = record[1]

                    for modifier in self._modifiers:
                        idx_dim = modifier[0]
                        idx_member = modifier[2][0]
                        idx_address[idx_dim] = idx_member

                    for key in values:
                        value = values[key]
                        if other._func:
                            value = other._func(value)
                        bolt = (0, tuple(idx_address), key)
                        self._cube._set(bolt, value)
                        # print(f"{idx_address} >>> {self._cube._idx_address_to_address(idx_address)}:= {value} is {self._cube._get(bolt)}")

            else:
                if not other._rows:
                    other.refresh()
                other.refresh()

                facts = other._cube._facts
                for row in other._rows:
                    idx_address = list(facts.addresses[row])
                    for modifier in self._modifiers:
                        idx_dim = modifier[0]
                        idx_member = modifier[2][0]
                        idx_address[idx_dim] = idx_member

                    for key in facts.facts[row]:
                        value = facts.facts[row][key]
                        if other._func:
                            value = other._func(value)
                        bolt = (0, tuple(idx_address), key)
                        self._cube._set(bolt, value)
                        # print(f"{idx_address} >>> {self._cube._idx_address_to_address(idx_address)}:= {value} is {self._cube._get(bolt)}")

    def identical_modifiers(self, modifiers_a, modifiers_b) -> bool:
        if len(modifiers_a) != len(modifiers_b):
            return False
        for a, b in zip(modifiers_a, modifiers_b):
            idx_dim_a, members_a, idx_members_a, level_members_a = a
            idx_dim_b, members_b, idx_members_b, level_members_b = b
            if idx_dim_a != idx_dim_b:
                return False
            if tuple(idx_members_a) != tuple(idx_members_b):
                return False
        return True

    def modifiers_of_same_scope(self, modifiers_a, modifiers_b) -> bool:
        if len(modifiers_a) == 0 or len(modifiers_b) == 0:
            return True
        elif len(modifiers_a) != len(modifiers_b):
            return False
        for a, b in zip(modifiers_a, modifiers_b):
            idx_dim_a, members_a, idx_members_a, level_members_a = a
            idx_dim_b, members_b, idx_members_b, level_members_b = b
            if idx_dim_a != idx_dim_b:
                return False
            if len(idx_members_a) != len(idx_members_b):
                return False
        return True

    def merge_modifier_into(self, modifiers_a, modifiers_b) -> bool:
        # todo: implement this...
        raise NotImplementedError()




# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
import string
import os

MEMBER_NAME_CHARS = set(string.ascii_letters + string.digits + '.-_/#+-*:,;|{}()"')
DB_OBJECT_NAME_CHARS = set(string.ascii_letters + string.digits + '_')


def is_valid_member_name(name: str):
    """Checks if a given name is valid (only contains characters by MEMBER_NAME_CHARS)."""
    return set(name) <= MEMBER_NAME_CHARS


def is_valid_db_object_name(name: str):
    """Checks if a given name is valid (only contains characters by DB_OBJECT_NAME_CHARS)."""
    return set(name) <= DB_OBJECT_NAME_CHARS


def to_valid_key(s: str):
    """Converts a string key to a valid TinyOlap object name."""
    s = str(s).strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)


def dict_keys_to_int(dictionary: dict):
    """Converts all string address in a dictionary containing numbers into integer address.
    Note: This is not a general solutions and sufficient for a special purpose."""
    converted = {}
    for k, v in dictionary.items():
        new_k = k
        if type(k) is str:
            if str(k).isdigit():
                new_k = int(k)
        if type(v) is dict:
            new_v = dict_keys_to_int(v)
        # elif not isinstance(v, str) and isinstance(v, collections.abc.Sequence):
        else:
            new_v = v
        converted[new_k] = new_v
    return converted


def get_path(folder: str = "db", create_path: bool = True):
    """
    Returns the path to a folder relative to the current working directory.
    :param folder: Name of the folder.
    :param create_path: Flag if the folder should be created if it does not exist.
    :return: The path of the folder.
    """
    folder = os.path.normpath(folder)
    directory = os.path.join(os.getcwd(), folder)
    if create_path:
        if not os.path.exists(directory):
            os.makedirs(directory)
    return directory


def get_file_path(file_name: str, folder: str = "db", create_path: bool = True):
    """
    Returns the path to a file and folder relative to the current working directory.
    :param file_name: Name of the file (incl. extensions).
    :param folder: Name of the folder.
    :param create_path: Flag if the folder should be created if it does not exist.
    :return: The path of the file.
    """
    file_folder = get_path(folder, create_path)
    return os.path.join(file_folder, file_name)



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in
# the LICENSE file in the root directory of this source tree.

from __future__ import annotations

import collections.abc
import json

from storage.storageprovider import StorageProvider
from tinyolap.member import Member
from tinyolap.case_insensitive_dict import CaseInsensitiveDict
from tinyolap.exceptions import *
from tinyolap.utils import *


class Dimension:
    """
    Dimensions are used to define the axis of a multi-dimensional :ref:`cube <cubes>`.
    Dimensions contain a list or hierarchy of :ref:`members <members>`.
    Members are string address and representing the entities of the dimension. e.g.,
    for a dimension called 'months' this would be the month names 'Jan' to 'Dec' or
    month aggregations like quarters or semesters.

    .. important::
        It is not foreseen to create *Dimension* objects directly! Always use the ``add_dimension()``
        method of a :ref:`database <databases>` to create a new dimension instance,
        as shown in the code fragment below.

        .. code:: python

            from tinyolap.database import Database

            # setup a new database
            database = Database("foo")

            # define dimensions
            dimension = db.add_dimension("bar")

    Within a single database, dimensions need to be unique by there name.
    """

    class MemberIndexManager:
        """
        Manages the available member index numbers.
        """

        def __init__(self, current: list[int] = None, first_free: int = 1):
            self.free: set[int] = set()
            self.next_new = max(1, first_free)
            if current:
                all_members = set(range(max(current) + 1))
                used_members = set(current)
                self.free: all_members - used_members
                self.next_new = max(current) + 1

        def clear(self):
            self.free: set[int] = set()
            self.next_new = 1

        def pop(self) -> int:
            if self.free:
                return self.free.pop()
            index = self.next_new
            self.next_new += 1
            return index

        def push(self, index):
            if type(index) is int:
                self.free.add(index)
            elif isinstance(index, collections.abc.Sequence):
                self.free.update(index)
            else:
                raise ValueError()

    __magic_key = object()

    @classmethod
    def _create(cls, storage_provider: StorageProvider, name: str, description: str = ""):
        """
        NOT INTENDED FOR EXTERNAL USE! Creates a new dimension.

        :param storage_provider: The storage provider of the database.
        :param name: Name of the dimension to be created.
        :param description: Description of the dimension to be added.
        :return: The new dimension.
        """
        dimension = Dimension(Dimension.__magic_key, name, description)
        dimension._storage_provider = storage_provider
        if storage_provider and storage_provider.connected:
            storage_provider.add_dimension(name, dimension.to_json())
        return dimension

    IDX = 0
    NAME = 1
    DESC = 2
    PARENTS = 3
    CHILDREN = 4
    ALL_PARENTS = 5
    LEVEL = 6
    ATTRIBUTES = 7
    BASE_CHILDREN = 8
    ALIASES = 9
    FORMAT = 10

    MEMBERS = 3
    IDX_MEMBERS = 4

    def __init__(self, dim_creation_key, name: str, description: str = ""):
        """
        NOT INTENDED FOR DIRECT USE! Cubes and dimensions always need to be managed by a Database.
        Use method 'Database.add_cube(...)' to create objects type Cube.

        :param dim_creation_key:
        :param name:
        :param description:
        """
        assert (dim_creation_key == Dimension.__magic_key), \
            "Objects of type Dimension can only be created through the method 'Database.add_dimension()'."

        self.name: str = name.strip()
        self.description: str = description
        self.members: dict[int, dict] = {}
        self._member_idx_manager = Dimension.MemberIndexManager()
        self._member_idx_lookup: CaseInsensitiveDict[str, int] = CaseInsensitiveDict()
        self._member_idx_list = []
        self.member_counter = 0
        self.highest_idx = 0

        self.database = None
        self._storage_provider: StorageProvider = None
        self.edit_mode: bool = False
        self.recovery_json = ""
        self.recovery_idx = set()

        self.alias_idx_lookup: CaseInsensitiveDict[str, int] = CaseInsensitiveDict()
        self.attributes: CaseInsensitiveDict[str, int] = CaseInsensitiveDict()
        self.attribute_query_caching: bool = True
        self.attribute_cache: CaseInsensitiveDict[str, list[str]] = CaseInsensitiveDict()
        self.subsets: CaseInsensitiveDict[str, dict] = CaseInsensitiveDict()
        self._subset_idx_manager = Dimension.MemberIndexManager()

    def __str__(self):
        """Returns the string representation of the dimension."""
        return f"dim:{self.name}"

    def __repr__(self):
        """Returns the string representation of the dimension."""
        return f"dim{self.name}"

    def __len__(self):
        """Returns the length (number of members) of the dimension"""
        return len(self.members)

    # region Dimension editing
    def clear(self) -> Dimension:
        """
        Deletes all members and all members from the dimension. Attributes will be kept.

        :return: The dimension itself.
        """
        self._member_idx_lookup = CaseInsensitiveDict()
        self.members = {}
        self._member_idx_manager.clear()
        self.alias_idx_lookup.clear()
        self.member_counter = 0
        self.subsets = {}
        self._subset_idx_manager.clear()
        return self

    def edit(self) -> Dimension:
        """
        Sets the dimension into edit mode. Required to add, remove or rename members,
        to add remove or edit subsets or attributes and alike.

        :return: The dimension itself.
        """
        if self.edit_mode:
            raise DimensionEditModeException("Failed to set edit mode. 'edit_begin()' was already called before.")
        self.edit_mode = True
        self.recovery_json = self.to_json()
        self.recovery_idx = set(self._member_idx_lookup.values())
        return self

    def commit(self) -> Dimension:
        """
        Commits all changes since 'edit_begin()' was called and ends the edit mode.

        :return: The dimension itself.
        """
        self.__update_member_hierarchies()
        if self._storage_provider and self._storage_provider.connected:
            self._storage_provider.add_dimension(self.name, self.to_json())

        # remove data for obsolete members (if any) from database
        obsolete = self.recovery_idx.difference(set(self._member_idx_lookup.values()))
        if obsolete:
            self.database._remove_members(self, obsolete)
        # update member list
        self._member_idx_list = [idx for idx in self.members.keys()]

        self.edit_mode = False
        self.database._flush_cache()
        return self

    def rollback(self) -> Dimension:
        """
        Rollback all changes since 'edit_begin()' was called and ends the edit mode.

        :return: The dimension itself.
        """
        self.from_json(self.recovery_json)
        self.edit_mode = False
        return self

    # endregion

    # region member context
    def member(self, member) -> Member:
        if type(member) is int:
            try:
                member_name = self.members[member]
                return Member(self, member_name,
                              None, idx_dim=-1,
                              idx_member=member,
                              member_level=self.members[self._member_idx_lookup[member_name]][self.LEVEL])
            except (IndexError, ValueError):
                raise KeyError(f"Failed to return Member with index '{member}'. The member does not exist.")

        elif member in self._member_idx_lookup:
            idx_member = self._member_idx_lookup[member]
            return Member(self, member,
                          None, idx_dim=-1,
                          idx_member=idx_member,
                          member_level=self.members[self._member_idx_lookup[member]][self.LEVEL])
        raise KeyError(f"Failed to return Member '{member}'. The member does not exist.")

    # region add, remove, rename members
    def add_member(self, member, children=None, description=None, number_format=None) -> Dimension:
        """Adds one or multiple members and (optionally) associated child-members to the dimension.

        :param member: A single string or an iterable of strings containing the members to be added.
        :param children: A single string or an iterable of strings containing the child members to be added.
               If parameter 'member' is an iterable of strings, then children must be an iterable of same size,
               either containing strings (adds a single child) or itself an iterable of string (adds multiple children).
        :param description: A description for the member to be added. If parameter 'member' is an iterable,
               then description will be ignored. For that case, please set descriptions for each member individually.
        :param number_format: A format string for output formatting, e.g. for numbers or percentages.
               Formatting follows the standard Python formatting specification at
               <https://docs.python.org/3/library/string.html#format-specification-mini-language>.
        :return Dimension: Returns the dimension itself.
        """
        if not self.edit_mode:
            raise DimensionEditModeException("Failed to add member. Dimension is not in edit mode.")

        member_list = member
        children_list = children
        multi = False

        if isinstance(member, str):
            if not self.__valid_member_name(member):
                raise KeyError(f"Failed to add member. Invalid member name '{member}'. "
                               f"'\\t', '\\n' and '\\r' characters are not supported.")

            member_list = [member]
            children_list = [children]
            multi = True
        if not children:
            children_list = [None] * len(member_list)

        for m, c in zip(member_list, children_list):
            # add the member
            idx_member = self.__member_add_parent_child(member=m, parent=None,
                                                        description=(None if multi else description))
            if number_format:
                self.members[idx_member][self.FORMAT] = number_format

            if c:
                # add children
                if isinstance(c, str):
                    c = [c]
                elif not (isinstance(c, collections.abc.Sequence) and not isinstance(c, str)):
                    raise DimensionEditModeException(
                        f"Failed to member '{m}' to dimension '{self.name}'. Unexpected type "
                        f"'{type(c)}' of parameter 'children' found.")
                for child in c:
                    if not isinstance(child, str):
                        raise DimensionEditModeException(
                            f"Failed to add child to member '{m}' of dimension '{self.name}. Unexpected type "
                            f"'{type(c)}' of parameter 'children' found.")
                    if not self.__valid_member_name(child):
                        raise KeyError(f"Failed to add member. Invalid member name '{child}'. "
                                       f"'\\t', '\\n' and '\\r' characters are not supported.")
                    self.__member_add_parent_child(member=child, parent=m, weight=1.0)

        return self

    def rename_member(self, member: str, new_name: str, new_description: str = None):
        """Renames a member."""
        if member not in self._member_idx_lookup:
            raise ValueError("Invalid or empty member name.")
        if not new_name:
            raise ValueError("Invalid or empty new member name.")
        if new_name == "*":
            raise ValueError("Invalid member new name. '*' is not a valid member name.")
        if new_name in self.members:
            raise ValueError("New name already exists.")

        idx_member = self._member_idx_lookup[member]
        self.members[idx_member][self.DESC] = (new_description if new_description else member)
        self._member_idx_lookup.pop(member)
        self._member_idx_lookup[new_name] = idx_member

        # adjust subsets
        for subset in self.subsets:
            if member in subset[self.MEMBERS]:
                idx = subset[self.MEMBERS].index(member)
                subset[self.MEMBERS][idx] = new_name

    def remove_member(self, member):
        """
        Removes one or multiple members from a dimension.

        :param member: The member or an iterable of members to be deleted.
        """
        member_list = member
        if isinstance(member, str):
            member_list = [member]

        # Ensure all members exist
        for member in member_list:
            if member not in self._member_idx_lookup:
                raise DimensionEditModeException(f"Failed to remove member(s). "
                                             f"At least 1 of {len(member_list)} member ('{member}') is not "
                                             f"a member of dimension {self.name}")

        # remove from directly related members
        for member in member_list:
            idx = self._member_idx_lookup[member]
            children = list(self.members[idx][self.CHILDREN])
            parents = list(self.members[idx][self.PARENTS])

            for child in children:
                if idx in self.members[child][self.PARENTS]:
                    self.members[child][self.PARENTS].remove(idx)
            for parent in parents:
                if idx in self.members[parent][self.CHILDREN]:
                    self.members[parent][self.CHILDREN].remove(idx)

        # remove from all related members
        for all_idx in self._member_idx_lookup.values():
            for member_idx in [self._member_idx_lookup[member] in member_list]:
                if member_idx in self.members[all_idx][self.ALL_PARENTS]:
                    self.members[all_idx][self.ALL_PARENTS].remove(member_idx)
                if member_idx in self.members[all_idx][self.BASE_CHILDREN]:
                    self.members[all_idx][self.BASE_CHILDREN].remove(member_idx)

        # remove the members
        member_idx_to_remove = [self._member_idx_lookup[m] for m in member_list]
        self._member_idx_manager.push(member_idx_to_remove)

        for m in member_list:
            del self._member_idx_lookup[m]
        for idx in member_idx_to_remove:
            if idx in self.members:
                del self.members[idx]

        # adjust subsets
        for subset in self.subsets.values():
            members_to_remove = set(member_list).intersection(set(subset[self.MEMBERS]))
            if members_to_remove:
                for member in members_to_remove:
                    idx = subset[self.MEMBERS].index(member)
                    subset[self.MEMBERS].pop(idx)
                    subset[self.IDX_MEMBERS].pop(idx)

    # endregion

    # region member aliases
    def member_add_alias(self, member: str, alias: str):
        """
        Adds a member alias to the dimension. Aliases enable the access of members
        by alternative names or address (e.g. a technical key, or an abbreviation).

        :param member: Name of the member to add an alias for.
        :param alias: The alias to be set.
        :raises KeyError: Raised if the member does not exist.
        :raises DuplicateKeyException: Raised if the alias is already used by another member.
                Individual aliases can only be assigned to one member.

        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"{member}' is not member a of dimension'{self.name}'")
        idx_member = self._member_idx_lookup[member]
        if alias in self.alias_idx_lookup:
            raise DuplicateKeyException(f"Duplicate alias. The alias '{alias}' is already used "
                                    f"by member '{self.members[idx_member][self.NAME]}' of dimension'{self.name}'")
        self.alias_idx_lookup[alias] = idx_member

    def remove_alias(self, alias: str):
        """
        Removes a member alias from the dimension.

        :param alias: The alias to be removed.
        """
        if alias not in self.alias_idx_lookup:
            raise KeyError(f"{alias}' is not alias a of dimension'{self.name}'")
        del self.alias_idx_lookup[alias]

    def member_remove_all_aliases(self, member: str):
        """
        Removes all aliases of a member from the dimension.

        :param member: Name of the member to remove the aliases for.
        :raises KeyError: Raised if the member does not exist.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"{member}' is not member a of dimension'{self.name}'")
        idx_member = self._member_idx_lookup[member]
        aliases_to_be_deleted = set([key for key, idx in self.alias_idx_lookup.items() if idx == idx_member])
        for alias in aliases_to_be_deleted:
            del self.alias_idx_lookup[alias]

    def member_has_alias(self, member: str) -> bool:
        """
        Checks if for a given member an alias is defined.

        :param member: Name of the member to be checked.
        :raises KeyError: Raised if the member does not exist.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"{member}' is not a member of dimension'{self.name}'")
        idx_member = self._member_idx_lookup[member]
        return idx_member in set(self.alias_idx_lookup.values())

    def member_aliases_count(self, member: str) -> int:
        """
        Returns the number of aliases defined for a given member.

        :param member: Name of the member to be checked.
        :raises KeyError: Raised if the member does not exist.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"Failed to return member alias count. '{member}' is not a member of dimension'{self.name}'")
        idx_member = self._member_idx_lookup[member]
        return len(set([idx for idx in set(self.alias_idx_lookup.values()) if idx == idx_member]))

    def get_member_by_alias(self, alias: str) -> str:
        """
        Returns the name of a member associated with the given.

        :param alias: Name of the alias to be checked.
        :raises KeyError: Raised if the alias does not exist.
        """
        if alias not in self.alias_idx_lookup:
            raise KeyError(f"Failed to get member by alias. '{alias}' is not a member alias of dimension'{self.name}'")
        idx_member = self.alias_idx_lookup[alias]
        return self.members[idx_member][self.NAME]

    # endregion

    # region member number_format (for output formatting)
    def member_set_format(self, member: str, format_string: str):
        """
        Set a number_format string for output formatting, especially useful for number formatting.
        Member formatting follows the standard Python formatting specification at
        https://docs.python.org/3/library/string.html#format-specification-mini-language.

        :param member: Name of the member to set the number_format for.
        :param format_string: The number_format string to be used. Member formatting follows the standard
               Python formatting specification at
               https://docs.python.org/3/library/string.html#format-specification-mini-language.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"Failed to set member number_format. '{member}' is not a member of dimension'{self.name}'")
        idx_member = self._member_idx_lookup[member]
        self.members[idx_member][self.FORMAT] = format_string

    def member_get_format(self, member: str) -> str:
        """
        Returns the number_format string of a member.

        :param member: Name of the member to return the number_format for.
        :return: Returns the number_format string for the member, or ``None`` if no number_format string is defined.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"Failed to return member number_format. '{member}' is not a member of dimension'{self.name}'")
        idx_member = self._member_idx_lookup[member]
        return self.members[idx_member][self.FORMAT]

    def member_remove_format(self, member: str):
        """
        Removes the number_format string of a member.

        :param member: Name of the member to remove the number_format for.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"Failed to remove member number_format. '{member}' is not a member of dimension'{self.name}'")
        idx_member = self._member_idx_lookup[member]
        self.members[idx_member][self.FORMAT] = None

    # endregion

    # region member information functions
    def member_get_ordinal(self, member: str) -> int:
        """
        Returns the ordinal position of a member with the list of members of the dimension.

        :param member: Name of the member to be checked.
        :return: The ordinal position of the member. If the member does not exits -1 will be returned
        """
        if member in self._member_idx_lookup:
            return list(self._member_idx_lookup.keys()).index(member)
        return -1

    def member_exists(self, member: str) -> bool:
        """
        Check if the member exists in the dimension.

        :param member: Name of the member to be checked.
        :return: ``True`` if the member exists, ``False`` otherwise.
        """
        return member in self._member_idx_lookup

    def member_get_index(self, member: str) -> int:
        """
        Returns the database internal index of a member.

        :param member: Name of the member to be evaluated.
        :return: The internal index of the member.
        :raises KeyError: Raised if the member does not exist.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"{member}' is not a member of dimension'{self.name}'")
        return self._member_idx_lookup[member]

    def member_get_parents(self, member: str) -> list[str]:
        """
        Returns a list of all parents of a member.
        :param member: Name of the member to be evaluated.
        :return: List of parents.
        :raises KeyError: Raised if the member does not exist.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"{member}' is not a member of dimension'{self.name}'")
        parents = []
        for idx in self.members[self._member_idx_lookup[member]][self.PARENTS]:
            parents.append(self.members[idx][self.NAME])
        return parents

    def member_get_children(self, member: str):
        """
        Returns a list of all children of a member.

        :param member: Name of the member to be evaluated.
        :return: List of children.
        :raises KeyError: Raised if the member does not exist.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"{member}' is not a member of dimension'{self.name}'")
        children = []
        for idx in self.members[self._member_idx_lookup[member]][self.CHILDREN]:
            children.append(self.members[idx][self.NAME])
        return children

    def member_get_leave_children(self, member: str):
        """
        Returns a list of all leave (base level) children (or grand children) of a member.

        :param member: Name of the member to be evaluated.
        :return: List of children.
        :raises KeyError: Raised if the member does not exist.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"{member}' is not a member of dimension'{self.name}'")
        if self.members[self._member_idx_lookup[member]][self.LEVEL] == 0:
            # already a base member, return that
            return [member, ]
        children = []
        for idx in self.members[self._member_idx_lookup[member]][self.CHILDREN]:
            if self.members[idx][self.LEVEL] > 0:
                members = self.member_get_leave_children(self.members[idx][self.NAME])
                children.extend(members)
            else:
                children.append(self.members[idx][self.NAME])
        return children


    def member_get_level(self, member: str):
        """
        Returns the level of a member within the member hierarchy.

        :param member: Name of the member to be evaluated.
        :return: 0 for leave level members, values > 0 for aggregated members.
        :raises KeyError: Raised if the member does not exist.
        """
        if member not in self._member_idx_lookup:
            raise KeyError(f"{member}' is not a member of dimension'{self.name}'")
        return self.members[self._member_idx_lookup[member]][self.LEVEL]

    # endregion

    # region member enumeration functions (returning lists of members)
    def get_members(self) -> list[str]:
        """
        Returns a list of all members of the dimension.

        :return: List of all members of the dimension.
        """
        return list(self.members[idx][self.NAME] for idx in self._member_idx_lookup.values())
        # return list(str(key) for key in self._member_idx_lookup.keys())

    def get_members_idx(self) -> list[int]:
        """
        Returns a list of indexes of all members of the dimension.

        :return: List of indexes of all members of the dimension.
        """
        return list(self._member_idx_lookup.values())

    def get_members_by_level(self, level: int) -> list[str]:
        """
        Returns a list of all members of the specific member level. 0 identifies the leave level of the dimension.

        :param level: Level of the members to be returned.
        :return: List of members of the specific member level.
        """
        return [self.members[idx_member][self.NAME]
                for idx_member in self.members
                if self.members[idx_member][self.LEVEL] == level]

    def get_top_level(self) -> int:
        """
        Returns the highest member level over all members of the dimension.

        :return: The highest member level over all members of the dimension.
        """
        return max([self.members[idx_member][self.LEVEL] for idx_member in self.members])

    def get_leave_members(self) -> list[str]:
        """
        Returns a list of all leave members (members without children = level equals 0) of the dimension.

        :return: List of leave level members of the dimension.
        """
        members = []
        for idx_member in self.members:
            if self.members[idx_member][self.LEVEL] == 0:
                members.append(self.members[idx_member][self.NAME])
        return members

    def get_aggregated_members(self) -> list[str]:
        """
        Returns a list of all aggregated members (members with children = level greater 0) of the dimension.

        :return: List of aggregated members of the dimension.
        """
        return [self.members[idx_member][self.NAME]
                for idx_member in self.members
                if self.members[idx_member][self.LEVEL] > 0]

    def get_root_members(self) -> list[str]:
        """
        Returns a list of all root members (members with a parent) of the dimension.

        :return: Returns a list of all root members of the dimension.
        """
        members = []
        for idx_member in self.members:
            if not self.members[idx_member][self.PARENTS]:
                members.append(self.members[idx_member][self.NAME])
        return members

    def get_first_member(self) -> str:
        """
        Returns the first member of the dimension.

        :return: Returns the first members of the dimension.
        """
        for idx_member in self.members:
            return self.members[idx_member][self.NAME]
        return None

    # endregion

    # region attributes
    def attributes_count(self) -> int:
        """
        Returns the number attributes defined in the dimension.

        :return: Number attributes defined in the dimension.
        """
        return len(self.attributes)

    def set_attribute(self, attribute: str, member: str, value):
        """
        Sets an attribute for a specific member of the dimension.

        :param member: Name of the member to set the attribute for.
        :param attribute: Name of the attribute to be set.
        :param value: Value to be set.
        :raises KeyError: Raised when either the member or the attribute name does not exist.
        :raises TypeError: Raised when the value if not of the expected type.
        """
        if attribute not in self.attributes:
            raise KeyError(f"Failed to set attribute value. "
                           f"'{attribute}' is not an attribute of dimension {self.name}.")
        expected_type = self.attributes[attribute]
        if not type(value) is expected_type:
            raise TypeError(f"Failed to set attribute value. "
                            f"Type of value is '{str(type(value))}' but '{str(expected_type)}' was expected.")
        if member not in self._member_idx_lookup:
            raise KeyError(f"Failed to set attribute value. "
                           f"'{member}' is not a member of dimension {self.name}.")
        idx = self._member_idx_lookup[member]
        self.members[idx][self.ATTRIBUTES][attribute] = value

        if self.attribute_query_caching:
            key = attribute + ":" + str(value)
            if key in self.attribute_cache:
                del self.attribute_cache[key]

    def get_attribute(self, attribute: str, member: str):
        """
        Returns the attribute value for a specific member of the dimension.

        :param attribute: Name of the attribute to be returned.
        :param member: Name of the member to get the attribute for.
        :raises KeyError: Raised when either the member or the attribute name does not exist.
        :return: The value of the attribute, or ``None`` if the attribute is not defined for the specific member.
        """
        if attribute not in self.attributes:
            raise KeyError(f"Failed to set attribute value. "
                           f"'{attribute}' is not an attribute of dimension {self.name}.")
        if member not in self._member_idx_lookup:
            raise KeyError(f"Failed to set attribute value. "
                           f"'{member}' is not a member of dimension {self.name}.")
        idx = self._member_idx_lookup[member]
        if attribute not in self.members[idx][self.ATTRIBUTES]:
            return None
        return self.members[idx][self.ATTRIBUTES][attribute]

    def get_attribute_type(self, attribute: str):
        """
        Returns the data type of an attribute.

        :param attribute: Name of the attribute to be returned.
        :raises KeyError: Raised when the attribute name does not exist.
        :return: The type of the attribute.
        """
        if attribute not in self.attributes:
            raise KeyError(f"Failed to set attribute value. "
                           f"'{attribute}' is not an attribute of dimension {self.name}.")
        return self.attributes[attribute]

    def has_attribute(self, attribute: str):
        """
        Checks if a specific attribute is defined for the dimension.

        :param attribute: Name of the attribute to be checked.
        :return: ``True``if the attribute exists. ``False`` otherwise.
        """
        return attribute in self.attributes

    def del_attribute_value(self, attribute: str, member: str):
        """
        Deletes an attribute value for a specific member of the dimension.

        :param attribute: Name of the attribute to be deleted.
        :param member: Name of the member to delete the attribute for.
        :raises KeyError: Raised when either the member or the attribute name does not exist.
        """
        if attribute not in self.attributes:
            raise KeyError(f"Failed to set attribute value. "
                           f"'{attribute}' is not an attribute of dimension {self.name}.")
        if member not in self._member_idx_lookup:
            raise KeyError(f"Failed to set attribute value. "
                           f"'{member}' is not a member of dimension {self.name}.")
        idx = self._member_idx_lookup[member]
        if attribute in self.members[idx][self.ATTRIBUTES]:
            del (self.members[idx][self.ATTRIBUTES][attribute])

    def add_attribute(self, attribute_name: str, value_type: type = object):
        """
        Adds an attribute field to the dimension. Attributes enable to store additional information
        along side of dimension members. Attributes have a value_type which is checked when an attribute
        value is set.

        :param attribute_name: Name of the attribute to be added.
        :param value_type: Type of value expected for the attribute. Default value is ``object`` to allow any data.
        :raises InvalidKeyException: Raised when the name of the attribute is invalid.
        :raises DuplicateKeyException: Raised when the name of the attribute already exists.
        """
        if not is_valid_db_object_name(attribute_name):
            raise InvalidKeyException(f"'{attribute_name}' is not a valid dimension attribute name. "
                                  f"Lower case alphanumeric characters and underscore supported only, "
                                  f"no whitespaces, no special characters.")
        if attribute_name in self.attributes:
            raise DuplicateKeyException(f"Failed to add attribute to dimension. "
                                    f"A dimension attribute named '{attribute_name}' already exists.")
        self.attributes[attribute_name] = value_type

    def rename_attribute(self, attribute_name: str, new_attribute_name: str):
        """
        Renames an attribute of a dimension.

        :param attribute_name: The name of the attribute to be renamed.
        :param new_attribute_name: The new name of the attribute.
        :raises InvalidKeyException: Raised when the new name of the attribute is invalid.
        :raises DuplicateKeyException: Raised when the new name of the attribute already exists.
        """
        if not is_valid_db_object_name(new_attribute_name):
            raise InvalidKeyException(f"Failed to rename dimension attribute. "
                                  f"'{new_attribute_name}' is not a valid dimension attribute name. "
                                  f"Lower case alphanumeric characters and underscore supported only, "
                                  f"no whitespaces, no special characters.")
        if attribute_name not in self.attributes:
            raise KeyError(f"Failed to rename dimension attribute. "
                           f"A dimension attribute named '{attribute_name}' does not exist.")

        # add new , remove old attribute values
        for member in self.members.values():
            if attribute_name in member[self.ATTRIBUTES]:
                member[self.ATTRIBUTES][new_attribute_name] = member[self.ATTRIBUTES][attribute_name]
                del (member[self.ATTRIBUTES][attribute_name])

    def remove_attribute(self, attribute_name: str):
        """
        Removes an attribute from the dimension.

        :param attribute_name: Name of the attribute to be removed.
        :raises KeyError: Raises KeyError if the attribute name not exists.

        """
        if attribute_name not in self.attributes:
            raise KeyError(f"Failed to remove attribute from dimension. "
                           f"A dimension attribute named '{attribute_name}' does not exist.")
        # delete all values
        for member in self.members.values():
            if attribute_name in member[self.ATTRIBUTES]:
                del (member[self.ATTRIBUTES][attribute_name])
        del (self.attributes[attribute_name])

    def get_members_by_attribute(self, attribute_name: str, attribute_value) -> list[str]:
        """
        Returns all members having a specific attribute value.

        :param attribute_name: Name of the attribute to be analyzed.
        :param attribute_value: Value of the attribute to used for filtering.
        :return:
        """
        if self.attribute_query_caching:
            key = attribute_name + ":" + str(attribute_value)
            if key in self.attribute_cache:
                return self.attribute_cache[key]

        if attribute_name not in self.attributes:
            raise KeyError(f"Failed to return members by attribute. "
                           f"'{attribute_name}' is not an attribute of dimension {self.name}.")
        members = []
        for idx_member in self.members:
            if attribute_name in self.members[idx_member][self.ATTRIBUTES]:
                if self.members[idx_member][self.ATTRIBUTES][attribute_name] == attribute_value:
                    members.append(self.members[idx_member][self.NAME])
        if self.attribute_query_caching:
            key = attribute_name + ":" + str(attribute_value)
            self.attribute_cache[key] = members
        return members

    # endregion

    # region subsets
    def add_subset(self, subset_name: str, members):
        """
        Adds a new subset to the dimension. A subset is a plain list of members,
        useful for calculation and reporting purposes.

        :param subset_name: Name of the subset to be added.
        :param members: A list (iterable) containing the member to be added to the subset.
        :raises InvalidKeyException: Raised when the name of the subset is invalid.
        :raises DuplicateKeyException: Raised when the name of the subset already exists.
        :raises TypeError: Raised when members list is not of the expected type (list or tuple)
        :raises KeyError: Raised when a member from the members list is not contained in the dimension.
        """
        if not is_valid_db_object_name(subset_name):
            raise InvalidKeyException(f"'{subset_name}' is not a valid dimension subset name. "
                                  f"Lower case alphanumeric characters and underscore supported only, "
                                  f"no whitespaces, no special characters.")
        if subset_name in self.subsets:
            raise DuplicateKeyException(f"Failed to add subset to dimension. "
                                    f"A dimension subset named '{subset_name}' already exists.")

        # validate members list
        if not ((type(members) is list) or (type(members) is tuple)):
            raise TypeError(f"Failed to add members to subset '{subset_name}'. "
                            f"Argument 'members' is not of expected type list or tuple, "
                            f"but of type '{type(subset_name)}'.")
        idx_members = []
        for member in members:
            if member in self._member_idx_lookup:
                idx_members.append(self._member_idx_lookup[member])
            else:
                raise KeyError(f"Failed to add member to subset. "
                               f"'{member}' is not a member of dimension {self.name}.")

        # create and add subset
        self.subsets[subset_name] = {self.IDX: self._subset_idx_manager.pop(),
                                     self.NAME: subset_name,
                                     self.MEMBERS: list(members),
                                     self.IDX_MEMBERS: idx_members}

    def has_subset(self, subset_name: str) -> bool:
        """
        Checks if a specific subset is defined for the dimension.

        :param subset_name: Name of the subset to be checked.
        :return: ``True``if the subset exists. ``False`` otherwise.
        """
        return subset_name in self.subsets

    def subsets_count(self) -> int:
        """
        Returns the number subsets defined in the dimension.

        :return: Number subsets defined in the dimension.
        """
        return len(self.subsets)

    def subset_contains(self, subset_name: str, member_name: str) -> bool:
        """
        Checks if a specific member is contained in a subset of the dimension.

        :param subset_name: Name of the subset to be checked.
        :param member_name: Name of the member to be checked.
        :return: ``True``if the member is contained in the subset. ``False`` otherwise.
        """
        if not subset_name in self.subsets:
            raise KeyError(f"Failed to check member contained in subset. "
                           f"'{subset_name}' is not a subset of dimension {self.name}.")
        return member_name in self.subsets[subset_name][self.MEMBERS]

    def rename_subset(self, subset_name: str, new_subset_name: str):
        """
        Renames a subset of the dimension.

        :param subset_name: Name of the subset to be added.
        :param new_subset_name: New name of the subset.
        :raises InvalidKeyException: Raised when the new name for the subset is invalid.
        :raises KeyError: Raised when the subset is not contained in the dimension.
        """
        if not is_valid_db_object_name(new_subset_name):
            raise InvalidKeyException(f"'{new_subset_name}' is not a valid dimension subset name. "
                                  f"Lower case alphanumeric characters and underscore supported only, "
                                  f"no whitespaces, no special characters.")
        if not subset_name in self.subsets:
            raise KeyError(f"Failed to rename subset. "
                           f"'{subset_name}' is not a subset of dimension {self.name}.")

        subset = self.subsets[subset_name]
        del self.subsets[subset_name]
        self.subsets[new_subset_name] = subset

    def get_subset(self, subset_name: str) -> tuple[str]:
        """
        Returns the list of member from a subset to the dimension.

        :param subset_name: Name of the subset to be return.
        :raises KeyError: Raised when the subset is not contained in the dimension.
        """
        if subset_name in self.subsets:
            return self.subsets[subset_name][self.MEMBERS]

        raise KeyError(f"Failed to return list of subset member. "
                       f"'{subset_name}' is not a subset of dimension {self.name}.")

    def remove_subset(self, subset_name: str):
        """
        Removes a subset from the dimension.

        :param subset_name: Name of the subset to be removed.
        :raises KeyError: Raised when the subset is not contained in the dimension.
        """
        if subset_name in self.subsets:
            self._subset_idx_manager.push(self.subsets[subset_name][self.IDX])
            del (self.subsets[subset_name])
            return

        raise KeyError(f"Failed to remove subset. "
                       f"'{subset_name}' is not a subset of dimension {self.name}.")

    # endregion

    # region serialization
    def to_json(self, beautify: bool = False):
        """
        Returns the json representation of the dimension. Helpful for serialization
        and deserialization of a dimension. The json returned by this function is
        the same as the one contained in the SQLite database storage provider (if available).

        :param beautify: Identifies if the json code should be beautified (multiple rows + indentation).
        :return: A json string representing the dimension.
        """
        data = ['{', f'"content": "dimension",', f'"name": "{self.name}",', f'"description": "{self.description}",',
                f'"count": {self.member_counter},', f'"members": {json.dumps(self.members)},',
                f'"lookup": {json.dumps(self._member_idx_lookup)},', f'"attributes": {json.dumps(self.attributes)},',
                f'"subsets": {json.dumps(self.subsets)}', '}']
        json_string = ''.join(data)
        if beautify:
            parsed = json.loads(json_string)
            json_string = json.dumps(parsed, indent=4)
        return json_string

    def from_json(self, json_string: str):
        """
        (Re-)initializes the dimension from a json string.

        .. warning::
            Calling this method for dimensions which are already used by cubes
            will very likely **corrupt your database!** Calling this method is only save
            **before** you create any cube. Handle with care.

        :param json_string: The json string containing the dimension definition.
        :raises FatalException: Raised if an error occurred during the deserialization from json string.
        """
        if not self.edit_mode:
            self.edit()

        try:
            # first, read everything
            dim = json.loads(json_string)
            new_name = dim["name"]
            new_description = dim["description"]
            new_count = dim["count"]
            new_members = dim["members"]
            new_member_idx_lookup = dim["lookup"]
            new_attributes = dim["attributes"]
            new_subsets = dim["subsets"]

            # json does not allow non-string address, but we use integer address. Conversion is required.
            new_members = dict_keys_to_int(new_members)

            # second, apply everything (this should not fail)
            self.name = new_name
            self.description = new_description
            self.member_counter = new_count
            self.members = new_members
            self._member_idx_lookup = CaseInsensitiveDict().populate(new_member_idx_lookup)
            self.attributes = new_attributes
            self.subsets = new_subsets
            self.commit()
        except Exception as err:
            raise FatalException(f"Failed to load json for dimension '{self.name}'. {str(err)}")

    # endregion

    # region auxiliary function to add, remove or rename members
    @staticmethod
    def __valid_member_name(name):
        return not (("\t" in name) or ("\n" in name) or ("\r" in name))

    def __member_add_parent_child(self, member, parent, weight: float = 1.0, description: str = None) -> int:
        if member in self._member_idx_lookup:
            member_idx = self._member_idx_lookup[member]
            if description:
                self.members[member_idx][self.DESC] = description
            if parent:
                self.__add_parent(member, parent)
        else:
            self.member_counter += 1
            member_idx = self._member_idx_manager.pop()
            self._member_idx_lookup[member] = member_idx
            self.members[member_idx] = {self.IDX: member_idx,
                                        self.NAME: member,
                                        self.DESC: (description if description else member),
                                        self.PARENTS: [],
                                        self.ALL_PARENTS: [],
                                        self.CHILDREN: [],
                                        self.LEVEL: 0,
                                        self.ATTRIBUTES: {},
                                        self.BASE_CHILDREN: [],
                                        self.ALIASES: [],
                                        self.FORMAT: None,
                                        }

            if parent:
                self.__add_parent(member, parent)
        return member_idx

    def __add_parent(self, member: str, parent: str = None):
        member_idx = self._member_idx_lookup[member]
        level = self.members[member_idx][self.LEVEL]
        if parent not in self._member_idx_lookup:
            # create parent member
            self.member_counter += 1
            parent_idx = self.member_counter
            self._member_idx_lookup[parent] = parent_idx
            self.members[parent_idx] = {self.IDX: parent_idx,
                                        self.NAME: parent,
                                        self.DESC: parent,
                                        self.PARENTS: [],
                                        self.ALL_PARENTS: [],
                                        self.CHILDREN: [member_idx],
                                        self.LEVEL: level + 1,
                                        self.ATTRIBUTES: {},
                                        self.BASE_CHILDREN: [],
                                        self.ALIASES: [],
                                        self.FORMAT: None,
                                        }
        else:
            parent_idx = self._member_idx_lookup[parent]
            self.members[parent_idx][self.LEVEL] = level + 1
            if member_idx not in self.members[parent_idx][self.CHILDREN]:
                self.members[parent_idx][self.CHILDREN].append(member_idx)

        # add new parent to member
        if parent_idx not in self.members[member_idx][self.PARENTS]:
            self.members[member_idx][self.PARENTS].append(parent_idx)

        # check for circular references
        if self.__circular_reference_detection(member_idx, member_idx):
            # remove the relationship
            self.members[member_idx][self.PARENTS].remove(parent_idx)
            self.members[parent_idx][self.CHILDREN].remove(member_idx)

            raise DimensionEditModeException(f"Circular reference detected on adding parent <-> child relation "
                                         f"'{self.members[parent_idx][self.NAME]}' <-> "
                                         f"'{self.members[member_idx][self.NAME]}' "
                                         f"to dimension {self.name}. Both members were added, "
                                         f"but the relation was not created.")

        # update all-parents list, only relevant for base level members
        self.__update_all_parents(member_idx, parent_idx)

    def __update_all_parents(self, idx, parent_idx):
        if self.members[idx][self.LEVEL] > 0:
            for child_idx in self.members[idx][self.CHILDREN]:
                self.__update_all_parents(child_idx, parent_idx)
        else:
            if parent_idx not in self.members[idx][self.ALL_PARENTS]:
                self.members[idx][self.ALL_PARENTS].append(parent_idx)

    def __update_member_hierarchies(self):
        for idx in self._member_idx_lookup.values():
            if self.members[idx][self.LEVEL] > 0:
                # update base level children
                self.members[idx][self.BASE_CHILDREN] = self.__get_base_members(idx)
            else:
                self.members[idx][self.ALL_PARENTS] = self.__get_all_parents(idx)

    def __check_circular_reference(self):
        for idx in self._member_idx_lookup.values():
            if self.__circular_reference_detection(idx, idx):
                raise DimensionEditModeException(f"Failed to commit dimension. Circular reference detected "
                                             f"for member {self.members[idx][self.NAME]}.")

    def __circular_reference_detection(self, start: int, current: int, visited=None):
        if visited is None:
            visited = set()

        if current in visited:
            return True

        visited.add(current)
        for parent in self.members[current][self.PARENTS]:
            if self.__circular_reference_detection(current, parent, visited):
                return True
        visited.remove(current)
        return False

    def __get_all_parents(self, idx) -> list[int]:
        all_parents = []
        for parent in self.members[idx][self.PARENTS]:
            all_parents.append(parent)
            all_parents = all_parents + self.__get_all_parents(parent)
        return all_parents

    def __get_base_members(self, idx) -> list[int]:
        if self.members[idx][self.LEVEL] == 0:
            return [idx]
        else:
            base_members = []
            for child_idx in self.members[idx][self.CHILDREN]:
                if self.members[child_idx][self.LEVEL] == 0:
                    base_members.append(child_idx)
                else:
                    base_members.extend(self.__get_base_members(child_idx))
            return base_members
    # endregion
